"""Test the reading of sensor_msgs messages."""
import math
from array import array
from pathlib import Path
from tempfile import TemporaryDirectory

import numpy as np
import pytest
from rosbags.rosbag2 import StoragePlugin, Writer
from rosbags.typesys import Stores, get_typestore
from rosbags.typesys.store import Typestore

from pybag.mcap_reader import McapFileReader
from pybag.schema.ros2msg import Ros2MsgError


def _find_mcap_file(temp_dir: str) -> Path:
    return next(Path(temp_dir).rglob('*.mcap'))


def _write_rosbags(
    temp_dir: str,
    msg,
    typestore,
    topic: str = '/rosbags',
    *,
    timestamp: int = 0,
) -> tuple[Path, int]:
    with Writer(Path(temp_dir) / 'rosbags', version=9, storage_plugin=StoragePlugin.MCAP) as writer:
        connection = writer.add_connection(topic, msg.__msgtype__, typestore=typestore)
        writer.write(connection, timestamp, typestore.serialize_cdr(msg, msg.__msgtype__))
    return _find_mcap_file(temp_dir), connection.id


def _make_header(typestore: Typestore, frame_id: str = "frame", sec: int = 1, nanosec: int = 2):
    Header = typestore.types["std_msgs/msg/Header"]
    Time = typestore.types["builtin_interfaces/msg/Time"]
    return Header(stamp=Time(sec=sec, nanosec=nanosec), frame_id=frame_id)


@pytest.fixture(params=[Stores.ROS2_JAZZY, Stores.ROS2_HUMBLE])
def typestore(request):
    return get_typestore(request.param)


def test_battery_state_rosbags(typestore: Typestore):
    BatteryState = typestore.types['sensor_msgs/msg/BatteryState']

    msg = BatteryState(
        header=_make_header(typestore),
        voltage=12.0,
        temperature=25.0,
        current=math.nan,
        charge=2.0,
        capacity=3.0,
        design_capacity=4.0,
        percentage=0.5,
        power_supply_status=BatteryState.POWER_SUPPLY_STATUS_DISCHARGING,
        power_supply_health=BatteryState.POWER_SUPPLY_HEALTH_GOOD,
        power_supply_technology=BatteryState.POWER_SUPPLY_TECHNOLOGY_LION,
        present=True,
        cell_voltage=np.array([3.5, 3.5, math.nan], dtype=np.float32),
        cell_temperature=np.array([30.0, 30.0, math.nan], dtype=np.float32),
        location='loc',
        serial_number='123',
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.voltage == 12.0
        assert messages[0].data.temperature == 25.0
        assert math.isnan(messages[0].data.current)
        assert messages[0].data.charge == 2.0
        assert messages[0].data.capacity == 3.0
        assert messages[0].data.design_capacity == 4.0
        assert messages[0].data.percentage == 0.5
        assert messages[0].data.power_supply_status == 2  # POWER_SUPPLY_STATUS_DISCHARGING
        assert messages[0].data.power_supply_health == 1  # POWER_SUPPLY_HEALTH_GOOD
        assert messages[0].data.power_supply_technology == 2  # POWER_SUPPLY_TECHNOLOGY_LION
        assert messages[0].data.present is True
        assert len(messages[0].data.cell_voltage) == 3
        assert messages[0].data.cell_voltage[0] == 3.5
        assert messages[0].data.cell_voltage[1] == 3.5
        assert math.isnan(messages[0].data.cell_voltage[2])
        assert len(messages[0].data.cell_temperature) == 3
        assert messages[0].data.cell_temperature[0] == 30.0
        assert messages[0].data.cell_temperature[1] == 30.0
        assert math.isnan(messages[0].data.cell_temperature[2])
        assert messages[0].data.location == 'loc'
        assert messages[0].data.serial_number == '123'


def test_camera_info_rosbags(typestore: Typestore):
    CameraInfo = typestore.types['sensor_msgs/msg/CameraInfo']
    RegionOfInterest = typestore.types['sensor_msgs/msg/RegionOfInterest']

    msg = CameraInfo(
        header=_make_header(typestore),
        height=480,
        width=640,
        distortion_model='plumb_bob',
        d=np.array([0.1, 0.2, 0.3, 0.4, 0.5], dtype=np.float64),
        k=np.array([float(i) for i in range(9)], dtype=np.float64),
        r=np.array([float(i) for i in range(9)], dtype=np.float64),
        p=np.array([float(i) for i in range(12)], dtype=np.float64),
        binning_x=1,
        binning_y=1,
        roi=RegionOfInterest(x_offset=0, y_offset=0, height=0, width=0, do_rectify=False),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.width == 640
        assert messages[0].data.height == 480
        assert messages[0].data.distortion_model == 'plumb_bob'
        assert messages[0].data.d == [0.1, 0.2, 0.3, 0.4, 0.5]
        assert messages[0].data.k == [float(i) for i in range(9)]
        assert messages[0].data.r == [float(i) for i in range(9)]
        assert messages[0].data.p == [float(i) for i in range(12)]
        assert messages[0].data.binning_x == 1
        assert messages[0].data.binning_y == 1
        assert messages[0].data.roi.x_offset == 0
        assert messages[0].data.roi.y_offset == 0
        assert messages[0].data.roi.height == 0
        assert messages[0].data.roi.width == 0
        assert messages[0].data.roi.do_rectify is False


def test_channel_float32_rosbags(typestore: Typestore):
    ChannelFloat32 = typestore.types['sensor_msgs/msg/ChannelFloat32']

    msg = ChannelFloat32(name='chan', values=np.array([1.0, 2.0], dtype=np.float32))
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.name == 'chan'
        assert messages[0].data.values == [1.0, 2.0]


def test_compressed_image_rosbags(typestore: Typestore):
    CompressedImage = typestore.types['sensor_msgs/msg/CompressedImage']

    msg = CompressedImage(
        header=_make_header(typestore),
        format='jpeg',
        data=np.array([1, 2, 3], dtype=np.uint8),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.format == 'jpeg'
        assert messages[0].data.data == [1, 2, 3]


def test_fluid_pressure_rosbags(typestore: Typestore):
    FluidPressure = typestore.types['sensor_msgs/msg/FluidPressure']

    msg = FluidPressure(
        header=_make_header(typestore),
        fluid_pressure=1013.25,
        variance=0.1,
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.fluid_pressure == 1013.25
        assert messages[0].data.variance == 0.1


def test_illuminance_rosbags(typestore: Typestore):
    Illuminance = typestore.types['sensor_msgs/msg/Illuminance']

    msg = Illuminance(
        header=_make_header(typestore),
        illuminance=100.0,
        variance=0.1,
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.illuminance == 100.0
        assert messages[0].data.variance == 0.1


def test_image_rosbags(typestore: Typestore):
    Image = typestore.types['sensor_msgs/msg/Image']

    msg = Image(
        header=_make_header(typestore),
        height=480,
        width=640,
        encoding='rgb8',
        is_bigendian=0,
        step=1920,
        data=np.array([0, 1, 2, 3], dtype=np.uint8),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.width == 640
        assert messages[0].data.height == 480
        assert messages[0].data.encoding == 'rgb8'
        assert messages[0].data.is_bigendian == 0
        assert messages[0].data.step == 1920
        assert messages[0].data.data == [0, 1, 2, 3]


def test_imu_rosbags(typestore: Typestore):
    Imu = typestore.types['sensor_msgs/msg/Imu']
    Quaternion = typestore.types['geometry_msgs/msg/Quaternion']
    Vector3 = typestore.types['geometry_msgs/msg/Vector3']

    msg = Imu(
        header=_make_header(typestore),
        orientation=Quaternion(x=0.0, y=0.0, z=0.0, w=1.0),
        orientation_covariance=np.array([float(i) for i in range(9)]),
        angular_velocity=Vector3(x=1.0, y=2.0, z=3.0),
        angular_velocity_covariance=np.array([float(i) for i in range(9)]),
        linear_acceleration=Vector3(x=4.0, y=5.0, z=6.0),
        linear_acceleration_covariance=np.array([float(i) for i in range(9)]),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.orientation.x == 0.0
        assert messages[0].data.orientation.y == 0.0
        assert messages[0].data.orientation.z == 0.0
        assert messages[0].data.orientation.w == 1.0
        assert messages[0].data.orientation_covariance == [float(i) for i in range(9)]
        assert messages[0].data.angular_velocity.x == 1.0
        assert messages[0].data.angular_velocity.y == 2.0
        assert messages[0].data.angular_velocity.z == 3.0
        assert messages[0].data.angular_velocity_covariance == [float(i) for i in range(9)]
        assert messages[0].data.linear_acceleration.x == 4.0
        assert messages[0].data.linear_acceleration.y == 5.0
        assert messages[0].data.linear_acceleration.z == 6.0
        assert messages[0].data.linear_acceleration_covariance == [float(i) for i in range(9)]


def test_joint_state_rosbags(typestore: Typestore):
    JointState = typestore.types['sensor_msgs/msg/JointState']

    msg = JointState(
        header=_make_header(typestore),
        name=['joint1', 'joint2'],
        position=np.array([1.0, 2.0], dtype=np.float64),
        velocity=np.array([3.0, 4.0], dtype=np.float64),
        effort=np.array([5.0, 6.0], dtype=np.float64),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.name == ['joint1', 'joint2']
        assert messages[0].data.position == [1.0, 2.0]
        assert messages[0].data.velocity == [3.0, 4.0]
        assert messages[0].data.effort == [5.0, 6.0]


def test_joy_rosbags(typestore: Typestore):
    Joy = typestore.types['sensor_msgs/msg/Joy']

    msg = Joy(
        header=_make_header(typestore),
        axes=np.array([1.0, 2.0], dtype=np.float32),
        buttons=np.array([1, 0], dtype=np.int32),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.axes == [1.0, 2.0]
        assert messages[0].data.buttons == [1, 0]


def test_joy_feedback_rosbags(typestore: Typestore):
    JoyFeedback = typestore.types['sensor_msgs/msg/JoyFeedback']

    msg = JoyFeedback(type=JoyFeedback.TYPE_LED, id=1, intensity=0.5)
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].channel_id == channel_id
        assert messages[0].data.type == JoyFeedback.TYPE_LED
        assert messages[0].data.id == 1
        assert messages[0].data.intensity == 0.5


def test_joy_feedback_array_rosbags(typestore: Typestore):
    JoyFeedbackArray = typestore.types['sensor_msgs/msg/JoyFeedbackArray']
    JoyFeedback = typestore.types['sensor_msgs/msg/JoyFeedback']

    msg = JoyFeedbackArray(
        array=[JoyFeedback(type=JoyFeedback.TYPE_LED, id=1, intensity=0.5)]
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert len(messages[0].data.array) == 1
        assert messages[0].data.array[0].type == 0  # TYPE_LED
        assert messages[0].data.array[0].id == 1
        assert messages[0].data.array[0].intensity == 0.5


def test_laser_echo_rosbags(typestore: Typestore):
    LaserEcho = typestore.types['sensor_msgs/msg/LaserEcho']

    msg = LaserEcho(echoes=np.array([1.0, 2.0], dtype=np.float32))
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.echoes == [1.0, 2.0]


def test_laser_scan_rosbags(typestore: Typestore):
    LaserScan = typestore.types['sensor_msgs/msg/LaserScan']

    msg = LaserScan(
        header=_make_header(typestore),
        angle_min=0.5,
        angle_max=0.5,
        angle_increment=0.5,
        time_increment=0.5,
        scan_time=0.5,
        range_min=0.5,
        range_max=10.0,
        ranges=np.array([1.0, 2.0], dtype=np.float32),
        intensities=np.array([3.0, 4.0], dtype=np.float32),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.angle_min == 0.5
        assert messages[0].data.angle_max == 0.5
        assert messages[0].data.angle_increment == 0.5
        assert messages[0].data.time_increment == 0.5
        assert messages[0].data.scan_time == 0.5
        assert messages[0].data.range_min == 0.5
        assert messages[0].data.range_max == 10.0
        assert messages[0].data.ranges == [1.0, 2.0]
        assert messages[0].data.intensities == [3.0, 4.0]


def test_magnetic_field_rosbags(typestore: Typestore):
    MagneticField = typestore.types['sensor_msgs/msg/MagneticField']
    Vector3 = typestore.types['geometry_msgs/msg/Vector3']

    msg = MagneticField(
        header=_make_header(typestore),
        magnetic_field=Vector3(x=1.0, y=2.0, z=3.0),
        magnetic_field_covariance=np.array([float(i) for i in range(9)]),
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.magnetic_field.x == 1.0
        assert messages[0].data.magnetic_field.y == 2.0
        assert messages[0].data.magnetic_field.z == 3.0
        assert messages[0].data.magnetic_field_covariance == [float(i) for i in range(9)]


def test_multi_dof_joint_state_rosbags(typestore: Typestore):
    MultiDOFJointState = typestore.types['sensor_msgs/msg/MultiDOFJointState']
    Transform = typestore.types['geometry_msgs/msg/Transform']
    Vector3 = typestore.types['geometry_msgs/msg/Vector3']
    Quaternion = typestore.types['geometry_msgs/msg/Quaternion']
    Twist = typestore.types['geometry_msgs/msg/Twist']
    Wrench = typestore.types['geometry_msgs/msg/Wrench']

    msg = MultiDOFJointState(
        header=_make_header(typestore),
        joint_names=['joint1'],
        transforms=[
            Transform(
                translation=Vector3(x=1.0, y=2.0, z=3.0),
                rotation=Quaternion(x=0.0, y=0.0, z=0.0, w=1.0),
            )
        ],
        twist=[
            Twist(
                linear=Vector3(x=1.0, y=2.0, z=3.0),
                angular=Vector3(x=4.0, y=5.0, z=6.0),
            )
        ],
        wrench=[
            Wrench(
                force=Vector3(x=1.0, y=2.0, z=3.0),
                torque=Vector3(x=4.0, y=5.0, z=6.0),
            )
        ],
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.joint_names == ['joint1']
        assert messages[0].data.transforms[0].translation.x == 1.0
        assert messages[0].data.transforms[0].translation.y == 2.0
        assert messages[0].data.transforms[0].translation.z == 3.0
        assert messages[0].data.transforms[0].rotation.x == 0.0
        assert messages[0].data.transforms[0].rotation.y == 0.0
        assert messages[0].data.transforms[0].rotation.z == 0.0
        assert messages[0].data.transforms[0].rotation.w == 1.0
        assert messages[0].data.twist[0].linear.x == 1.0
        assert messages[0].data.twist[0].linear.y == 2.0
        assert messages[0].data.twist[0].linear.z == 3.0
        assert messages[0].data.twist[0].angular.x == 4.0
        assert messages[0].data.twist[0].angular.y == 5.0
        assert messages[0].data.twist[0].angular.z == 6.0
        assert messages[0].data.wrench[0].force.x == 1.0
        assert messages[0].data.wrench[0].force.y == 2.0
        assert messages[0].data.wrench[0].force.z == 3.0
        assert messages[0].data.wrench[0].torque.x == 4.0
        assert messages[0].data.wrench[0].torque.y == 5.0
        assert messages[0].data.wrench[0].torque.z == 6.0


def test_multi_echo_laser_scan_rosbags(typestore: Typestore):
    MultiEchoLaserScan = typestore.types['sensor_msgs/msg/MultiEchoLaserScan']
    LaserEcho = typestore.types['sensor_msgs/msg/LaserEcho']

    msg = MultiEchoLaserScan(
        header=_make_header(typestore),
        angle_min=0.5,
        angle_max=0.5,
        angle_increment=0.5,
        time_increment=0.5,
        scan_time=0.5,
        range_min=0.5,
        range_max=10.0,
        ranges=[LaserEcho(echoes=np.array([1.0, 2.0], dtype=np.float32))],
        intensities=[LaserEcho(echoes=np.array([3.0, 4.0], dtype=np.float32))],
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.angle_min == 0.5
        assert messages[0].data.angle_max == 0.5
        assert messages[0].data.angle_increment == 0.5
        assert messages[0].data.time_increment == 0.5
        assert messages[0].data.scan_time == 0.5
        assert messages[0].data.range_min == 0.5
        assert messages[0].data.range_max == 10.0
        assert messages[0].data.ranges[0].echoes == [1.0, 2.0]
        assert messages[0].data.intensities[0].echoes == [3.0, 4.0]


def test_nav_sat_fix_rosbags(typestore: Typestore):
    NavSatFix = typestore.types['sensor_msgs/msg/NavSatFix']
    NavSatStatus = typestore.types['sensor_msgs/msg/NavSatStatus']

    msg = NavSatFix(
        header=_make_header(typestore),
        status=NavSatStatus(status=NavSatStatus.STATUS_FIX, service=NavSatStatus.SERVICE_GPS),
        latitude=1.0,
        longitude=2.0,
        altitude=3.0,
        position_covariance=np.array([float(i) for i in range(9)]),
        position_covariance_type=NavSatFix.COVARIANCE_TYPE_DIAGONAL_KNOWN,
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.status.status == 0  # STATUS_FIX
        assert messages[0].data.latitude == 1.0
        assert messages[0].data.longitude == 2.0
        assert messages[0].data.altitude == 3.0
        assert messages[0].data.position_covariance == [float(i) for i in range(9)]
        assert messages[0].data.position_covariance_type == 2  # COVARIANCE_TYPE_DIAGONAL_KNOWN


def test_nav_sat_status_rosbags(typestore: Typestore):
    NavSatStatus = typestore.types['sensor_msgs/msg/NavSatStatus']

    msg = NavSatStatus(status=NavSatStatus.STATUS_FIX, service=NavSatStatus.SERVICE_GPS)
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.status == 0  # STATUS_FIX
        assert messages[0].data.service == 1  # SERVICE_GPS


def test_point_cloud2_rosbags(typestore: Typestore):
    PointCloud2 = typestore.types['sensor_msgs/msg/PointCloud2']
    PointField = typestore.types['sensor_msgs/msg/PointField']

    msg = PointCloud2(
        header=_make_header(typestore),
        height=1,
        width=2,
        fields=[PointField(name='x', offset=0, datatype=PointField.FLOAT32, count=1)],
        is_bigendian=False,
        point_step=4,
        row_step=8,
        data=np.array([1, 2, 3, 4, 5, 6, 7, 8], dtype=np.uint8),
        is_dense=True,
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.height == 1
        assert messages[0].data.width == 2
        assert messages[0].data.fields[0].name == 'x'
        assert messages[0].data.fields[0].offset == 0
        assert messages[0].data.fields[0].datatype == 7  # FLOAT32
        assert messages[0].data.fields[0].count == 1
        assert messages[0].data.is_bigendian == False
        assert messages[0].data.point_step == 4
        assert messages[0].data.row_step == 8
        assert messages[0].data.data == [1, 2, 3, 4, 5, 6, 7, 8]
        assert messages[0].data.is_dense == True


def test_point_field_rosbags(typestore: Typestore):
    PointField = typestore.types['sensor_msgs/msg/PointField']

    msg = PointField(name='x', offset=0, datatype=PointField.FLOAT32, count=1)
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.name == 'x'
        assert messages[0].data.offset == 0
        assert messages[0].data.datatype == 7  # FLOAT32
        assert messages[0].data.count == 1


def test_range_rosbags(typestore: Typestore):
    Range = typestore.types['sensor_msgs/msg/Range']

    kwargs = dict(
        header=_make_header(typestore),
        radiation_type=Range.ULTRASOUND,
        field_of_view=0.5,
        min_range=0.2,
        max_range=10.0,
        range=5.0,
    )
    # variance was added after ROS 2 Humble
    if hasattr(Range, '__dataclass_fields__') and 'variance' in Range.__dataclass_fields__:
        kwargs['variance'] = 0.5
    msg = Range(**kwargs)
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.radiation_type == 0  # ULTRASOUND
        assert messages[0].data.range == 5.0
        if hasattr(Range, '__dataclass_fields__') and 'variance' in Range.__dataclass_fields__:
            assert messages[0].data.variance == 0.5


def test_region_of_interest_rosbags(typestore: Typestore):
    RegionOfInterest = typestore.types['sensor_msgs/msg/RegionOfInterest']

    msg = RegionOfInterest(x_offset=1, y_offset=2, height=3, width=4, do_rectify=True)
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.x_offset == 1
        assert messages[0].data.y_offset == 2
        assert messages[0].data.height == 3
        assert messages[0].data.width == 4
        assert messages[0].data.do_rectify is True


def test_relative_humidity_rosbags(typestore: Typestore):
    RelativeHumidity = typestore.types['sensor_msgs/msg/RelativeHumidity']

    msg = RelativeHumidity(
        header=_make_header(typestore),
        relative_humidity=0.5,
        variance=0.1,
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.relative_humidity == 0.5
        assert messages[0].data.variance == 0.1


def test_temperature_rosbags(typestore: Typestore):
    Temperature = typestore.types['sensor_msgs/msg/Temperature']

    msg = Temperature(
        header=_make_header(typestore),
        temperature=36.5,
        variance=0.1,
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].channel_id == channel_id
        assert messages[0].data.temperature == 36.5
        assert messages[0].data.variance == 0.1


def test_time_reference_rosbags(typestore: Typestore):
    TimeReference = typestore.types['sensor_msgs/msg/TimeReference']
    Time = typestore.types['builtin_interfaces/msg/Time']

    msg = TimeReference(
        header=_make_header(typestore),
        time_ref=Time(sec=3, nanosec=4),
        source='source',
    )
    with TemporaryDirectory() as temp_dir:
        mcap_file, channel_id = _write_rosbags(temp_dir, msg, typestore)
        reader = McapFileReader.from_file(mcap_file)
        messages = list(reader.messages('/rosbags'))

        assert len(messages) == 1
        assert messages[0].log_time == 0
        assert messages[0].publish_time == 0
        assert messages[0].sequence == 0
        assert messages[0].channel_id == channel_id
        assert messages[0].data.header.stamp.sec == 1
        assert messages[0].data.header.stamp.nanosec == 2
        assert messages[0].data.header.frame_id == 'frame'
        assert messages[0].data.time_ref.sec == 3
        assert messages[0].data.time_ref.nanosec == 4
        assert messages[0].data.source == 'source'



from dataclasses import dataclass

import pybag.types as t


@dataclass(kw_only=True)
class Duration:
    __msg_name__ = 'builtin_interfaces/msg/Duration'

    sec: t.int32
    nanosec: t.uint32


@dataclass(kw_only=True)
class Time:
    __msg_name__ = 'builtin_interfaces/msg/Time'

    sec: t.int32
    nanosec: t.uint32



"""Test reading nav_msgs messages written with the official MCAP writer."""

from pathlib import Path
from tempfile import TemporaryDirectory
from textwrap import dedent
from typing import Any

from mcap_ros2.writer import Writer as McapWriter

from pybag.mcap_reader import McapFileReader


def _write_mcap(temp_dir: str, msg: Any, msgtype: str, schema_text: str) -> Path:
    path = Path(temp_dir) / "test.mcap"
    with open(path, "wb") as f:
        writer = McapWriter(f)
        schema = writer.register_msgdef(msgtype, schema_text)
        writer.write_message(
            topic="/rosbags",
            schema=schema,
            message=msg,
            log_time=0,
            publish_time=0,
            sequence=0,
        )
        writer.finish()
    return path


def test_nav_msgs_goals():
    msgtype = "nav_msgs/Goals"
    schema = dedent("""
        # An array of poses that represents goals for a robot
        std_msgs/Header header
        geometry_msgs/PoseStamped[] goals
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.
        builtin_interfaces/Time stamp
        string frame_id
        ================================================================================
        MSG: geometry_msgs/PoseStamped
        # A Pose with reference coordinate frame and timestamp
        std_msgs/Header header
        Pose pose
        ================================================================================
        MSG: geometry_msgs/Pose
        # A representation of position and orientation in free space
        Point position
        Quaternion orientation
        ================================================================================
        MSG: geometry_msgs/Point
        # This contains the position of a point in free space
        float64 x
        float64 y
        float64 z
        ================================================================================
        MSG: geometry_msgs/Quaternion
        # This represents an orientation in free space in quaternion form.
        float64 x 0
        float64 y 0
        float64 z 0
        float64 w 1
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "map"
            },
            "goals": [
                {
                    "header": {
                        "stamp": {"sec": 124, "nanosec": 100000},
                        "frame_id": "map"
                    },
                    "pose": {
                        "position": {"x": 1.0, "y": 2.0, "z": 3.0},
                        "orientation": {"x": 0.0, "y": 0.0, "z": 0.0, "w": 1.0}
                    }
                }
            ]
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "map"
    assert len(messages[0].data.goals) == 1
    assert messages[0].data.goals[0].header.stamp.sec == 124
    assert messages[0].data.goals[0].header.stamp.nanosec == 100000
    assert messages[0].data.goals[0].header.frame_id == "map"
    assert messages[0].data.goals[0].pose.position.x == 1.0
    assert messages[0].data.goals[0].pose.position.y == 2.0
    assert messages[0].data.goals[0].pose.position.z == 3.0
    assert messages[0].data.goals[0].pose.orientation.x == 0.0
    assert messages[0].data.goals[0].pose.orientation.y == 0.0
    assert messages[0].data.goals[0].pose.orientation.z == 0.0
    assert messages[0].data.goals[0].pose.orientation.w == 1.0


def test_nav_msgs_grid_cells():
    msgtype = "nav_msgs/GridCells"
    schema = dedent("""
        # An array of cells in a 2D grid
        std_msgs/Header header
        float32 cell_width
        float32 cell_height
        geometry_msgs/Point[] cells
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.
        builtin_interfaces/Time stamp
        string frame_id
        ================================================================================
        MSG: geometry_msgs/Point
        # This contains the position of a point in free space
        float64 x
        float64 y
        float64 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "map"
            },
            "cell_width": 0.05,
            "cell_height": 0.05,
            "cells": [
                {"x": 1.0, "y": 2.0, "z": 0.0},
                {"x": 1.5, "y": 2.5, "z": 0.0}
            ]
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "map"
    assert abs(messages[0].data.cell_width - 0.05) < 0.001
    assert abs(messages[0].data.cell_height - 0.05) < 0.001
    assert len(messages[0].data.cells) == 2
    assert messages[0].data.cells[0].x == 1.0
    assert messages[0].data.cells[0].y == 2.0
    assert messages[0].data.cells[0].z == 0.0
    assert messages[0].data.cells[1].x == 1.5
    assert messages[0].data.cells[1].y == 2.5
    assert messages[0].data.cells[1].z == 0.0


def test_nav_msgs_map_meta_data():
    msgtype = "nav_msgs/MapMetaData"
    schema = dedent("""
        # This hold basic information about the characterists of the OccupancyGrid

        # The time at which the map was loaded
        builtin_interfaces/Time map_load_time
        # The map resolution [m/cell]
        float32 resolution
        # Map width [cells]
        uint32 width
        # Map height [cells]
        uint32 height
        # The origin of the map [m, m, rad]. This is the real-world pose of the
        # cell (0,0) in the map.
        geometry_msgs/Pose origin
        ================================================================================
        MSG: geometry_msgs/Pose
        # A representation of position and orientation in free space
        Point position
        Quaternion orientation
        ================================================================================
        MSG: geometry_msgs/Point
        # This contains the position of a point in free space
        float64 x
        float64 y
        float64 z
        ================================================================================
        MSG: geometry_msgs/Quaternion
        # This represents an orientation in free space in quaternion form.
        float64 x 0
        float64 y 0
        float64 z 0
        float64 w 1
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "map_load_time": {"sec": 100, "nanosec": 500000},
            "resolution": 0.05,
            "width": 800,
            "height": 600,
            "origin": {
                "position": {"x": -20.0, "y": -15.0, "z": 0.0},
                "orientation": {"x": 0.0, "y": 0.0, "z": 0.0, "w": 1.0}
            }
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.map_load_time.sec == 100
    assert messages[0].data.map_load_time.nanosec == 500000
    assert abs(messages[0].data.resolution - 0.05) < 0.001
    assert messages[0].data.width == 800
    assert messages[0].data.height == 600
    assert messages[0].data.origin.position.x == -20.0
    assert messages[0].data.origin.position.y == -15.0
    assert messages[0].data.origin.position.z == 0.0
    assert messages[0].data.origin.orientation.x == 0.0
    assert messages[0].data.origin.orientation.y == 0.0
    assert messages[0].data.origin.orientation.z == 0.0
    assert messages[0].data.origin.orientation.w == 1.0


def test_nav_msgs_occupancy_grid():
    msgtype = "nav_msgs/OccupancyGrid"
    schema = dedent("""
        # This represents a 2-D grid map, in which each cell represents the probability of
        # occupancy.

        std_msgs/Header header

        # MetaData for the map
        nav_msgs/MapMetaData info

        # The map data, in row-major order, starting with (0,0). Occupancy
        # probabilities are in the range [0,100]. Unknown is -1.
        int8[] data
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.
        builtin_interfaces/Time stamp
        string frame_id
        ================================================================================
        MSG: nav_msgs/MapMetaData
        # This hold basic information about the characterists of the OccupancyGrid

        # The time at which the map was loaded
        builtin_interfaces/Time map_load_time
        # The map resolution [m/cell]
        float32 resolution
        # Map width [cells]
        uint32 width
        # Map height [cells]
        uint32 height
        # The origin of the map [m, m, rad]. This is the real-world pose of the
        # cell (0,0) in the map.
        geometry_msgs/Pose origin
        ================================================================================
        MSG: geometry_msgs/Pose
        # A representation of position and orientation in free space
        Point position
        Quaternion orientation
        ================================================================================
        MSG: geometry_msgs/Point
        # This contains the position of a point in free space
        float64 x
        float64 y
        float64 z
        ================================================================================
        MSG: geometry_msgs/Quaternion
        # This represents an orientation in free space in quaternion form.
        float64 x 0
        float64 y 0
        float64 z 0
        float64 w 1
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "map"
            },
            "info": {
                "map_load_time": {"sec": 100, "nanosec": 500000},
                "resolution": 0.25,
                "width": 4,
                "height": 3,
                "origin": {
                    "position": {"x": 0.0, "y": 0.0, "z": 0.0},
                    "orientation": {"x": 0.0, "y": 0.0, "z": 0.0, "w": 1.0}
                }
            },
            "data": [0, -1, 100, 50, 25, 75, 0, -1, 100, 50, 25, 75]  # 4x3 grid
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "map"
    assert messages[0].data.info.map_load_time.sec == 100
    assert messages[0].data.info.map_load_time.nanosec == 500000
    assert messages[0].data.info.resolution == 0.25
    assert messages[0].data.info.width == 4
    assert messages[0].data.info.height == 3
    assert messages[0].data.info.origin.position.x == 0.0
    assert messages[0].data.info.origin.position.y == 0.0
    assert messages[0].data.info.origin.position.z == 0.0
    assert messages[0].data.info.origin.orientation.x == 0.0
    assert messages[0].data.info.origin.orientation.y == 0.0
    assert messages[0].data.info.origin.orientation.z == 0.0
    assert messages[0].data.info.origin.orientation.w == 1.0
    assert messages[0].data.data == [0, -1, 100, 50, 25, 75, 0, -1, 100, 50, 25, 75]


def test_nav_msgs_odometry():
    msgtype = "nav_msgs/Odometry"
    schema = dedent("""
        # This represents an estimate of a position and velocity in free space.
        # The pose in this message should be specified in the coordinate frame given by header.frame_id.
        # The twist in this message should be specified in the coordinate frame given by the child_frame_id
        std_msgs/Header header
        string child_frame_id
        geometry_msgs/PoseWithCovariance pose
        geometry_msgs/TwistWithCovariance twist
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.
        builtin_interfaces/Time stamp
        string frame_id
        ================================================================================
        MSG: geometry_msgs/PoseWithCovariance
        # This represents a pose in free space with uncertainty.

        Pose pose

        # Row-major representation of the 6x6 covariance matrix
        # The orientation parameters use a fixed-axis representation.
        # In order, the parameters are:
        # (x, y, z, rotation about X axis, rotation about Y axis, rotation about Z axis)
        float64[36] covariance
        ================================================================================
        MSG: geometry_msgs/Pose
        # A representation of position and orientation in free space
        Point position
        Quaternion orientation
        ================================================================================
        MSG: geometry_msgs/Point
        # This contains the position of a point in free space
        float64 x
        float64 y
        float64 z
        ================================================================================
        MSG: geometry_msgs/Quaternion
        # This represents an orientation in free space in quaternion form.
        float64 x 0
        float64 y 0
        float64 z 0
        float64 w 1
        ================================================================================
        MSG: geometry_msgs/TwistWithCovariance
        # This expresses velocity in free space with uncertainty.

        Twist twist

        # Row-major representation of the 6x6 covariance matrix
        # The orientation parameters use a fixed-axis representation.
        # In order, the parameters are:
        # (x, y, z, rotation about X axis, rotation about Y axis, rotation about Z axis)
        float64[36] covariance
        ================================================================================
        MSG: geometry_msgs/Twist
        # This expresses velocity in free space broken into its linear and angular parts.
        Vector3 linear
        Vector3 angular
        ================================================================================
        MSG: geometry_msgs/Vector3
        # This represents a vector in free space.
        # It is only meant to represent a direction. Therefore, it does not
        # make sense to apply a translation to it (e.g., when applying a
        # generic rigid transformation to a Vector3, tf2 will only apply the
        # rotation). If you want your data to be translatable too, use the
        # geometry_msgs/Point message instead.

        float64 x
        float64 y
        float64 z
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "odom"
            },
            "child_frame_id": "base_link",
            "pose": {
                "pose": {
                    "position": {"x": 1.0, "y": 2.0, "z": 0.0},
                    "orientation": {"x": 0.0, "y": 0.0, "z": 0.0, "w": 1.0}
                },
                "covariance": [0.0] * 36
            },
            "twist": {
                "twist": {
                    "linear": {"x": 0.5, "y": 0.0, "z": 0.0},
                    "angular": {"x": 0.0, "y": 0.0, "z": 0.1}
                },
                "covariance": [0.0] * 36
            }
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "odom"
    assert messages[0].data.child_frame_id == "base_link"
    assert messages[0].data.pose.pose.position.x == 1.0
    assert messages[0].data.pose.pose.position.y == 2.0
    assert messages[0].data.pose.pose.position.z == 0.0
    assert messages[0].data.pose.pose.orientation.x == 0.0
    assert messages[0].data.pose.pose.orientation.y == 0.0
    assert messages[0].data.pose.pose.orientation.z == 0.0
    assert messages[0].data.pose.pose.orientation.w == 1.0
    assert messages[0].data.pose.covariance == [0.0] * 36
    assert messages[0].data.twist.twist.linear.x == 0.5
    assert messages[0].data.twist.twist.linear.y == 0.0
    assert messages[0].data.twist.twist.linear.z == 0.0
    assert messages[0].data.twist.twist.angular.x == 0.0
    assert messages[0].data.twist.twist.angular.y == 0.0
    assert messages[0].data.twist.twist.angular.z == 0.1
    assert messages[0].data.twist.covariance == [0.0] * 36


def test_nav_msgs_path():
    msgtype = "nav_msgs/Path"
    schema = dedent("""
        # An array of poses that represents a Path for a robot to follow
        std_msgs/Header header
        geometry_msgs/PoseStamped[] poses
        ================================================================================
        MSG: std_msgs/Header
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.
        builtin_interfaces/Time stamp
        string frame_id
        ================================================================================
        MSG: geometry_msgs/PoseStamped
        # A Pose with reference coordinate frame and timestamp
        std_msgs/Header header
        Pose pose
        ================================================================================
        MSG: geometry_msgs/Pose
        # A representation of position and orientation in free space
        Point position
        Quaternion orientation
        ================================================================================
        MSG: geometry_msgs/Point
        # This contains the position of a point in free space
        float64 x
        float64 y
        float64 z
        ================================================================================
        MSG: geometry_msgs/Quaternion
        # This represents an orientation in free space in quaternion form.
        float64 x 0
        float64 y 0
        float64 z 0
        float64 w 1
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "header": {
                "stamp": {"sec": 123, "nanosec": 456789},
                "frame_id": "map"
            },
            "poses": [
                {
                    "header": {
                        "stamp": {"sec": 124, "nanosec": 100000},
                        "frame_id": "map"
                    },
                    "pose": {
                        "position": {"x": 0.0, "y": 0.0, "z": 0.0},
                        "orientation": {"x": 0.0, "y": 0.0, "z": 0.0, "w": 1.0}
                    }
                },
            ]
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.header.stamp.sec == 123
    assert messages[0].data.header.stamp.nanosec == 456789
    assert messages[0].data.header.frame_id == "map"
    assert len(messages[0].data.poses) == 1
    assert messages[0].data.poses[0].header.stamp.sec == 124
    assert messages[0].data.poses[0].header.stamp.nanosec == 100000
    assert messages[0].data.poses[0].header.frame_id == "map"
    assert messages[0].data.poses[0].pose.position.x == 0.0
    assert messages[0].data.poses[0].pose.position.y == 0.0
    assert messages[0].data.poses[0].pose.position.z == 0.0
    assert messages[0].data.poses[0].pose.orientation.x == 0.0
    assert messages[0].data.poses[0].pose.orientation.y == 0.0
    assert messages[0].data.poses[0].pose.orientation.z == 0.0
    assert messages[0].data.poses[0].pose.orientation.w == 1.0



"""Test reading std_msgs messages written with the official MCAP writer."""

from pathlib import Path
from tempfile import TemporaryDirectory
from textwrap import dedent
from typing import Any

from mcap_ros2.writer import Writer as McapWriter

from pybag.mcap_reader import McapFileReader


def _write_mcap(temp_dir: str, msg: Any, msgtype: str, schema_text: str) -> Path:
    path = Path(temp_dir) / "test.mcap"
    with open(path, "wb") as f:
        writer = McapWriter(f)
        schema = writer.register_msgdef(msgtype, schema_text)
        writer.write_message(
            topic="/rosbags",
            schema=schema,
            message=msg,
            log_time=0,
            publish_time=0,
            sequence=0,
        )
        writer.finish()
    return path


def test_std_msgs_bool():
    msgtype = "std_msgs/Bool"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        bool data
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"data": True}
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.data is True


def test_std_msgs_byte():
    msgtype = "std_msgs/Byte"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        byte data
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"data": 42}  # b'\x2a'
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.data == b'\x2a'


def test_std_msgs_byte_multi_array():
    msgtype = "std_msgs/ByteMultiArray"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # Please look at the MultiArrayLayout message definition for
        # documentation on all multiarrays.

        MultiArrayLayout  layout        # specification of data layout
        byte[]            data          # array of data
        ================================================================================
        MSG: std_msgs/MultiArrayLayout
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # The multiarray declares a generic multi-dimensional array of a
        # particular data type.  Dimensions are ordered from outer most
        # to inner most.
        #
        # Accessors should ALWAYS be written in terms of dimension stride
        # and specified outer-most dimension first.
        #
        # multiarray(i,j,k) = data[data_offset + dim_stride[1]*i + dim_stride[2]*j + k]
        #
        # A standard, 3-channel 640x480 image with interleaved color channels
        # would be specified as:
        #
        # dim[0].label  = "height"
        # dim[0].size   = 480
        # dim[0].stride = 3*640*480 = 921600  (note dim[0] stride is just size of image)
        # dim[1].label  = "width"
        # dim[1].size   = 640
        # dim[1].stride = 3*640 = 1920
        # dim[2].label  = "channel"
        # dim[2].size   = 3
        # dim[2].stride = 3
        #
        # multiarray(i,j,k) refers to the ith row, jth column, and kth channel.

        MultiArrayDimension[] dim # Array of dimension properties
        uint32 data_offset        # padding bytes at front of data
        ================================================================================
        MSG: std_msgs/MultiArrayDimension
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        string label   # label of given dimension
        uint32 size    # size of given dimension (in type units)
        uint32 stride  # stride of given dimension
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "layout": {
                "dim": [{"label": "x", "size": 3, "stride": 3}],
                "data_offset": 0
            },
            "data": [1, 2, 3]
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert len(messages[0].data.layout.dim) == 1
    assert messages[0].data.layout.dim[0].label == "x"
    assert messages[0].data.layout.dim[0].size == 3
    assert messages[0].data.layout.dim[0].stride == 3
    assert messages[0].data.layout.data_offset == 0
    assert list(messages[0].data.data) == [b'\x01', b'\x02', b'\x03']


def test_std_msgs_char():
    msgtype = "std_msgs/Char"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        char data
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"data": 65}  # ASCII 'A'
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.data == 'A'


def test_std_msgs_colorrgba():
    msgtype = "std_msgs/ColorRGBA"
    schema = dedent("""
        float32 r
        float32 g
        float32 b
        float32 a
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"r": 1.0, "g": 0.5, "b": 0.25, "a": 0.125}
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.r == 1.0
    assert messages[0].data.g == 0.5
    assert messages[0].data.b == 0.25
    assert messages[0].data.a == 0.125


def test_std_msgs_empty():
    msgtype = "std_msgs/Empty"
    schema = ""

    with TemporaryDirectory() as temp_dir:
        msg = {}
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    # Empty messages should return a dataclass instance, not None
    assert messages[0].data is not None
    assert hasattr(messages[0].data, '__msg_name__')
    assert messages[0].data.__msg_name__ == 'std_msgs/Empty'


def test_std_msgs_float32():
    msgtype = "std_msgs/Float32"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        float32 data
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"data": 3.5}
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.data == 3.5


def test_std_msgs_float32_multi_array():
    msgtype = "std_msgs/Float32MultiArray"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # Please look at the MultiArrayLayout message definition for
        # documentation on all multiarrays.

        MultiArrayLayout  layout        # specification of data layout
        float32[]         data          # array of data
        ================================================================================
        MSG: std_msgs/MultiArrayLayout
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # The multiarray declares a generic multi-dimensional array of a
        # particular data type.  Dimensions are ordered from outer most
        # to inner most.
        #
        # Accessors should ALWAYS be written in terms of dimension stride
        # and specified outer-most dimension first.
        #
        # multiarray(i,j,k) = data[data_offset + dim_stride[1]*i + dim_stride[2]*j + k]
        #
        # A standard, 3-channel 640x480 image with interleaved color channels
        # would be specified as:
        #
        # dim[0].label  = "height"
        # dim[0].size   = 480
        # dim[0].stride = 3*640*480 = 921600  (note dim[0] stride is just size of image)
        # dim[1].label  = "width"
        # dim[1].size   = 640
        # dim[1].stride = 3*640 = 1920
        # dim[2].label  = "channel"
        # dim[2].size   = 3
        # dim[2].stride = 3
        #
        # multiarray(i,j,k) refers to the ith row, jth column, and kth channel.

        MultiArrayDimension[] dim # Array of dimension properties
        uint32 data_offset        # padding bytes at front of data
        ================================================================================
        MSG: std_msgs/MultiArrayDimension
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        string label   # label of given dimension
        uint32 size    # size of given dimension (in type units)
        uint32 stride  # stride of given dimension
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "layout": {
                "dim": [{"label": "x", "size": 2, "stride": 2}],
                "data_offset": 0
            },
            "data": [1.5, 2.5]
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert len(messages[0].data.layout.dim) == 1
    assert messages[0].data.layout.dim[0].label == "x"
    assert messages[0].data.layout.dim[0].size == 2
    assert messages[0].data.layout.dim[0].stride == 2
    assert messages[0].data.layout.data_offset == 0
    assert messages[0].data.data[0] == 1.5
    assert messages[0].data.data[1] == 2.5


def test_std_msgs_float64():
    msgtype = "std_msgs/Float64"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        float64 data
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"data": 2.71828}
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.data == 2.71828


def test_std_msgs_float64_multi_array():
    msgtype = "std_msgs/Float64MultiArray"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # Please look at the MultiArrayLayout message definition for
        # documentation on all multiarrays.

        MultiArrayLayout  layout        # specification of data layout
        float64[]         data          # array of data
        ================================================================================
        MSG: std_msgs/MultiArrayLayout
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # The multiarray declares a generic multi-dimensional array of a
        # particular data type.  Dimensions are ordered from outer most
        # to inner most.
        #
        # Accessors should ALWAYS be written in terms of dimension stride
        # and specified outer-most dimension first.
        #
        # multiarray(i,j,k) = data[data_offset + dim_stride[1]*i + dim_stride[2]*j + k]
        #
        # A standard, 3-channel 640x480 image with interleaved color channels
        # would be specified as:
        #
        # dim[0].label  = "height"
        # dim[0].size   = 480
        # dim[0].stride = 3*640*480 = 921600  (note dim[0] stride is just size of image)
        # dim[1].label  = "width"
        # dim[1].size   = 640
        # dim[1].stride = 3*640 = 1920
        # dim[2].label  = "channel"
        # dim[2].size   = 3
        # dim[2].stride = 3
        #
        # multiarray(i,j,k) refers to the ith row, jth column, and kth channel.

        MultiArrayDimension[] dim # Array of dimension properties
        uint32 data_offset        # padding bytes at front of data
        ================================================================================
        MSG: std_msgs/MultiArrayDimension
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        string label   # label of given dimension
        uint32 size    # size of given dimension (in type units)
        uint32 stride  # stride of given dimension
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "layout": {
                "dim": [{"label": "x", "size": 2, "stride": 2}],
                "data_offset": 0
            },
            "data": [3.14159, 2.71828]
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert len(messages[0].data.layout.dim) == 1
    assert messages[0].data.layout.dim[0].label == "x"
    assert messages[0].data.layout.dim[0].size == 2
    assert messages[0].data.layout.dim[0].stride == 2
    assert messages[0].data.layout.data_offset == 0
    assert messages[0].data.data[0] == 3.14159
    assert messages[0].data.data[1] == 2.71828


def test_std_msgs_header():
    msgtype = "std_msgs/Header"
    schema = dedent("""
        # Standard metadata for higher-level stamped data types.
        # This is generally used to communicate timestamped data
        # in a particular coordinate frame.

        # Two-integer timestamp that is expressed as seconds and nanoseconds.
        builtin_interfaces/Time stamp

        # Transform frame with which this data is associated.
        string frame_id
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "stamp": {"sec": 123, "nanosec": 456789},
            "frame_id": "test_frame"
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.stamp.sec == 123
    assert messages[0].data.stamp.nanosec == 456789
    assert messages[0].data.frame_id == "test_frame"


def test_std_msgs_int16():
    msgtype = "std_msgs/Int16"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        int16 data
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"data": -1234}
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.data == -1234


def test_std_msgs_int16_multi_array():
    msgtype = "std_msgs/Int16MultiArray"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # Please look at the MultiArrayLayout message definition for
        # documentation on all multiarrays.

        MultiArrayLayout  layout        # specification of data layout
        int16[]           data          # array of data
        ================================================================================
        MSG: std_msgs/MultiArrayLayout
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # The multiarray declares a generic multi-dimensional array of a
        # particular data type.  Dimensions are ordered from outer most
        # to inner most.
        #
        # Accessors should ALWAYS be written in terms of dimension stride
        # and specified outer-most dimension first.
        #
        # multiarray(i,j,k) = data[data_offset + dim_stride[1]*i + dim_stride[2]*j + k]
        #
        # A standard, 3-channel 640x480 image with interleaved color channels
        # would be specified as:
        #
        # dim[0].label  = "height"
        # dim[0].size   = 480
        # dim[0].stride = 3*640*480 = 921600  (note dim[0] stride is just size of image)
        # dim[1].label  = "width"
        # dim[1].size   = 640
        # dim[1].stride = 3*640 = 1920
        # dim[2].label  = "channel"
        # dim[2].size   = 3
        # dim[2].stride = 3
        #
        # multiarray(i,j,k) refers to the ith row, jth column, and kth channel.

        MultiArrayDimension[] dim # Array of dimension properties
        uint32 data_offset        # padding bytes at front of data
        ================================================================================
        MSG: std_msgs/MultiArrayDimension
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        string label   # label of given dimension
        uint32 size    # size of given dimension (in type units)
        uint32 stride  # stride of given dimension
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "layout": {
                "dim": [{"label": "x", "size": 3, "stride": 3}],
                "data_offset": 0
            },
            "data": [-100, 0, 100]
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert len(messages[0].data.layout.dim) == 1
    assert messages[0].data.layout.dim[0].label == "x"
    assert messages[0].data.layout.dim[0].size == 3
    assert messages[0].data.layout.dim[0].stride == 3
    assert messages[0].data.layout.data_offset == 0
    assert list(messages[0].data.data) == [-100, 0, 100]


def test_std_msgs_int32():
    msgtype = "std_msgs/Int32"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        int32 data
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"data": -123456}
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.data == -123456


def test_std_msgs_int32_multi_array():
    msgtype = "std_msgs/Int32MultiArray"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # Please look at the MultiArrayLayout message definition for
        # documentation on all multiarrays.

        MultiArrayLayout  layout        # specification of data layout
        int32[]           data          # array of data
        ================================================================================
        MSG: std_msgs/MultiArrayLayout
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # The multiarray declares a generic multi-dimensional array of a
        # particular data type.  Dimensions are ordered from outer most
        # to inner most.
        #
        # Accessors should ALWAYS be written in terms of dimension stride
        # and specified outer-most dimension first.
        #
        # multiarray(i,j,k) = data[data_offset + dim_stride[1]*i + dim_stride[2]*j + k]
        #
        # A standard, 3-channel 640x480 image with interleaved color channels
        # would be specified as:
        #
        # dim[0].label  = "height"
        # dim[0].size   = 480
        # dim[0].stride = 3*640*480 = 921600  (note dim[0] stride is just size of image)
        # dim[1].label  = "width"
        # dim[1].size   = 640
        # dim[1].stride = 3*640 = 1920
        # dim[2].label  = "channel"
        # dim[2].size   = 3
        # dim[2].stride = 3
        #
        # multiarray(i,j,k) refers to the ith row, jth column, and kth channel.

        MultiArrayDimension[] dim # Array of dimension properties
        uint32 data_offset        # padding bytes at front of data
        ================================================================================
        MSG: std_msgs/MultiArrayDimension
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        string label   # label of given dimension
        uint32 size    # size of given dimension (in type units)
        uint32 stride  # stride of given dimension
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "layout": {
                "dim": [{"label": "x", "size": 2, "stride": 2}],
                "data_offset": 0
            },
            "data": [-1000, 2000]
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.layout.data_offset == 0
    assert len(messages[0].data.layout.dim) == 1
    assert messages[0].data.layout.dim[0].label == "x"
    assert messages[0].data.layout.dim[0].size == 2
    assert messages[0].data.layout.dim[0].stride == 2
    assert list(messages[0].data.data) == [-1000, 2000]


def test_std_msgs_int64():
    msgtype = "std_msgs/Int64"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        int64 data
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"data": -9876543210}
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.data == -9876543210


def test_std_msgs_int64_multi_array():
    msgtype = "std_msgs/Int64MultiArray"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # Please look at the MultiArrayLayout message definition for
        # documentation on all multiarrays.

        MultiArrayLayout  layout        # specification of data layout
        int64[]           data          # array of data
        ================================================================================
        MSG: std_msgs/MultiArrayLayout
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # The multiarray declares a generic multi-dimensional array of a
        # particular data type.  Dimensions are ordered from outer most
        # to inner most.
        #
        # Accessors should ALWAYS be written in terms of dimension stride
        # and specified outer-most dimension first.
        #
        # multiarray(i,j,k) = data[data_offset + dim_stride[1]*i + dim_stride[2]*j + k]
        #
        # A standard, 3-channel 640x480 image with interleaved color channels
        # would be specified as:
        #
        # dim[0].label  = "height"
        # dim[0].size   = 480
        # dim[0].stride = 3*640*480 = 921600  (note dim[0] stride is just size of image)
        # dim[1].label  = "width"
        # dim[1].size   = 640
        # dim[1].stride = 3*640 = 1920
        # dim[2].label  = "channel"
        # dim[2].size   = 3
        # dim[2].stride = 3
        #
        # multiarray(i,j,k) refers to the ith row, jth column, and kth channel.

        MultiArrayDimension[] dim # Array of dimension properties
        uint32 data_offset        # padding bytes at front of data
        ================================================================================
        MSG: std_msgs/MultiArrayDimension
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        string label   # label of given dimension
        uint32 size    # size of given dimension (in type units)
        uint32 stride  # stride of given dimension
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "layout": {
                "dim": [{"label": "x", "size": 2, "stride": 2}],
                "data_offset": 0
            },
            "data": [-9223372036854775807, 9223372036854775807]
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert len(messages[0].data.layout.dim) == 1
    assert messages[0].data.layout.dim[0].label == "x"
    assert messages[0].data.layout.dim[0].size == 2
    assert messages[0].data.layout.dim[0].stride == 2
    assert messages[0].data.layout.data_offset == 0
    assert list(messages[0].data.data) == [-9223372036854775807, 9223372036854775807]


def test_std_msgs_int8():
    msgtype = "std_msgs/Int8"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        int8 data
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"data": -42}
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.data == -42


def test_std_msgs_int8_multi_array():
    msgtype = "std_msgs/Int8MultiArray"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # Please look at the MultiArrayLayout message definition for
        # documentation on all multiarrays.

        MultiArrayLayout  layout        # specification of data layout
        int8[]            data          # array of data
        ================================================================================
        MSG: std_msgs/MultiArrayLayout
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # The multiarray declares a generic multi-dimensional array of a
        # particular data type.  Dimensions are ordered from outer most
        # to inner most.
        #
        # Accessors should ALWAYS be written in terms of dimension stride
        # and specified outer-most dimension first.
        #
        # multiarray(i,j,k) = data[data_offset + dim_stride[1]*i + dim_stride[2]*j + k]
        #
        # A standard, 3-channel 640x480 image with interleaved color channels
        # would be specified as:
        #
        # dim[0].label  = "height"
        # dim[0].size   = 480
        # dim[0].stride = 3*640*480 = 921600  (note dim[0] stride is just size of image)
        # dim[1].label  = "width"
        # dim[1].size   = 640
        # dim[1].stride = 3*640 = 1920
        # dim[2].label  = "channel"
        # dim[2].size   = 3
        # dim[2].stride = 3
        #
        # multiarray(i,j,k) refers to the ith row, jth column, and kth channel.

        MultiArrayDimension[] dim # Array of dimension properties
        uint32 data_offset        # padding bytes at front of data
        ================================================================================
        MSG: std_msgs/MultiArrayDimension
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        string label   # label of given dimension
        uint32 size    # size of given dimension (in type units)
        uint32 stride  # stride of given dimension
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "layout": {
                "dim": [{"label": "x", "size": 3, "stride": 3}],
                "data_offset": 0
            },
            "data": [-128, 0, 127]
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert len(messages[0].data.layout.dim) == 1
    assert messages[0].data.layout.dim[0].label == "x"
    assert messages[0].data.layout.dim[0].size == 3
    assert messages[0].data.layout.dim[0].stride == 3
    assert messages[0].data.layout.data_offset == 0
    assert list(messages[0].data.data) == [-128, 0, 127]


def test_std_msgs_multi_array_dimension():
    msgtype = "std_msgs/MultiArrayDimension"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        string label   # label of given dimension
        uint32 size    # size of given dimension (in type units)
        uint32 stride  # stride of given dimension
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "label": "x",
            "size": 10,
            "stride": 40
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.label == "x"
    assert messages[0].data.size == 10
    assert messages[0].data.stride == 40


def test_std_msgs_multi_array_layout():
    msgtype = "std_msgs/MultiArrayLayout"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # The multiarray declares a generic multi-dimensional array of a
        # particular data type.  Dimensions are ordered from outer most
        # to inner most.
        #
        # Accessors should ALWAYS be written in terms of dimension stride
        # and specified outer-most dimension first.
        #
        # multiarray(i,j,k) = data[data_offset + dim_stride[1]*i + dim_stride[2]*j + k]
        #
        # A standard, 3-channel 640x480 image with interleaved color channels
        # would be specified as:
        #
        # dim[0].label  = "height"
        # dim[0].size   = 480
        # dim[0].stride = 3*640*480 = 921600  (note dim[0] stride is just size of image)
        # dim[1].label  = "width"
        # dim[1].size   = 640
        # dim[1].stride = 3*640 = 1920
        # dim[2].label  = "channel"
        # dim[2].size   = 3
        # dim[2].stride = 3
        #
        # multiarray(i,j,k) refers to the ith row, jth column, and kth channel.

        MultiArrayDimension[] dim # Array of dimension properties
        uint32 data_offset        # padding bytes at front of data
        ================================================================================
        MSG: std_msgs/MultiArrayDimension
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        string label   # label of given dimension
        uint32 size    # size of given dimension (in type units)
        uint32 stride  # stride of given dimension
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "dim": [
                {"label": "x", "size": 10, "stride": 40},
                {"label": "y", "size": 4, "stride": 4}
            ],
            "data_offset": 0
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert len(messages[0].data.dim) == 2
    assert messages[0].data.dim[0].label == "x"
    assert messages[0].data.dim[0].size == 10
    assert messages[0].data.dim[0].stride == 40
    assert messages[0].data.dim[1].label == "y"
    assert messages[0].data.dim[1].size == 4
    assert messages[0].data.dim[1].stride == 4
    assert messages[0].data.data_offset == 0


def test_std_msgs_string():
    msgtype = "std_msgs/String"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        string data
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"data": "Hello, World!"}
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.data == "Hello, World!"


def test_std_msgs_uint16():
    msgtype = "std_msgs/UInt16"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        uint16 data
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"data": 1234}
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.data == 1234


def test_std_msgs_uint16_multi_array():
    msgtype = "std_msgs/UInt16MultiArray"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # Please look at the MultiArrayLayout message definition for
        # documentation on all multiarrays.

        MultiArrayLayout  layout        # specification of data layout
        uint16[]            data        # array of data
        ================================================================================
        MSG: std_msgs/MultiArrayLayout
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # The multiarray declares a generic multi-dimensional array of a
        # particular data type.  Dimensions are ordered from outer most
        # to inner most.
        #
        # Accessors should ALWAYS be written in terms of dimension stride
        # and specified outer-most dimension first.
        #
        # multiarray(i,j,k) = data[data_offset + dim_stride[1]*i + dim_stride[2]*j + k]
        #
        # A standard, 3-channel 640x480 image with interleaved color channels
        # would be specified as:
        #
        # dim[0].label  = "height"
        # dim[0].size   = 480
        # dim[0].stride = 3*640*480 = 921600  (note dim[0] stride is just size of image)
        # dim[1].label  = "width"
        # dim[1].size   = 640
        # dim[1].stride = 3*640 = 1920
        # dim[2].label  = "channel"
        # dim[2].size   = 3
        # dim[2].stride = 3
        #
        # multiarray(i,j,k) refers to the ith row, jth column, and kth channel.

        MultiArrayDimension[] dim # Array of dimension properties
        uint32 data_offset        # padding bytes at front of data
        ================================================================================
        MSG: std_msgs/MultiArrayDimension
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        string label   # label of given dimension
        uint32 size    # size of given dimension (in type units)
        uint32 stride  # stride of given dimension
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "layout": {
                "dim": [{"label": "x", "size": 2, "stride": 2}],
                "data_offset": 0
            },
            "data": [0, 65535]
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert len(messages[0].data.layout.dim) == 1
    assert messages[0].data.layout.dim[0].label == "x"
    assert messages[0].data.layout.dim[0].size == 2
    assert messages[0].data.layout.dim[0].stride == 2
    assert messages[0].data.layout.data_offset == 0
    assert list(messages[0].data.data) == [0, 65535]


def test_std_msgs_uint32():
    msgtype = "std_msgs/UInt32"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        uint32 data
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"data": 123456}
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.data == 123456


def test_std_msgs_uint32_multi_array():
    msgtype = "std_msgs/UInt32MultiArray"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # Please look at the MultiArrayLayout message definition for
        # documentation on all multiarrays.

        MultiArrayLayout  layout        # specification of data layout
        uint32[]          data          # array of data
        ================================================================================
        MSG: std_msgs/MultiArrayLayout
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # The multiarray declares a generic multi-dimensional array of a
        # particular data type.  Dimensions are ordered from outer most
        # to inner most.
        #
        # Accessors should ALWAYS be written in terms of dimension stride
        # and specified outer-most dimension first.
        #
        # multiarray(i,j,k) = data[data_offset + dim_stride[1]*i + dim_stride[2]*j + k]
        #
        # A standard, 3-channel 640x480 image with interleaved color channels
        # would be specified as:
        #
        # dim[0].label  = "height"
        # dim[0].size   = 480
        # dim[0].stride = 3*640*480 = 921600  (note dim[0] stride is just size of image)
        # dim[1].label  = "width"
        # dim[1].size   = 640
        # dim[1].stride = 3*640 = 1920
        # dim[2].label  = "channel"
        # dim[2].size   = 3
        # dim[2].stride = 3
        #
        # multiarray(i,j,k) refers to the ith row, jth column, and kth channel.

        MultiArrayDimension[] dim # Array of dimension properties
        uint32 data_offset        # padding bytes at front of data
        ================================================================================
        MSG: std_msgs/MultiArrayDimension
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        string label   # label of given dimension
        uint32 size    # size of given dimension (in type units)
        uint32 stride  # stride of given dimension
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "layout": {
                "dim": [{"label": "x", "size": 2, "stride": 2}],
                "data_offset": 0
            },
            "data": [0, 4294967295]
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert len(messages[0].data.layout.dim) == 1
    assert messages[0].data.layout.dim[0].label == "x"
    assert messages[0].data.layout.dim[0].size == 2
    assert messages[0].data.layout.dim[0].stride == 2
    assert messages[0].data.layout.data_offset == 0
    assert list(messages[0].data.data) == [0, 4294967295]


def test_std_msgs_uint64():
    msgtype = "std_msgs/UInt64"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        uint64 data
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"data": 9876543210}
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.data == 9876543210


def test_std_msgs_uint64_multi_array():
    msgtype = "std_msgs/UInt64MultiArray"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # Please look at the MultiArrayLayout message definition for
        # documentation on all multiarrays.

        MultiArrayLayout  layout        # specification of data layout
        uint64[]          data          # array of data
        ================================================================================
        MSG: std_msgs/MultiArrayLayout
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # The multiarray declares a generic multi-dimensional array of a
        # particular data type.  Dimensions are ordered from outer most
        # to inner most.
        #
        # Accessors should ALWAYS be written in terms of dimension stride
        # and specified outer-most dimension first.
        #
        # multiarray(i,j,k) = data[data_offset + dim_stride[1]*i + dim_stride[2]*j + k]
        #
        # A standard, 3-channel 640x480 image with interleaved color channels
        # would be specified as:
        #
        # dim[0].label  = "height"
        # dim[0].size   = 480
        # dim[0].stride = 3*640*480 = 921600  (note dim[0] stride is just size of image)
        # dim[1].label  = "width"
        # dim[1].size   = 640
        # dim[1].stride = 3*640 = 1920
        # dim[2].label  = "channel"
        # dim[2].size   = 3
        # dim[2].stride = 3
        #
        # multiarray(i,j,k) refers to the ith row, jth column, and kth channel.

        MultiArrayDimension[] dim # Array of dimension properties
        uint32 data_offset        # padding bytes at front of data
        ================================================================================
        MSG: std_msgs/MultiArrayDimension
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        string label   # label of given dimension
        uint32 size    # size of given dimension (in type units)
        uint32 stride  # stride of given dimension
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "layout": {
                "dim": [{"label": "x", "size": 2, "stride": 2}],
                "data_offset": 0
            },
            "data": [0, 18446744073709551615]
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert len(messages[0].data.layout.dim) == 1
    assert messages[0].data.layout.dim[0].label == "x"
    assert messages[0].data.layout.dim[0].size == 2
    assert messages[0].data.layout.dim[0].stride == 2
    assert messages[0].data.layout.data_offset == 0
    assert list(messages[0].data.data) == [0, 18446744073709551615]


def test_std_msgs_uint8():
    msgtype = "std_msgs/UInt8"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        uint8 data
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {"data": 255}
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert messages[0].data.data == 255


def test_std_msgs_uint8_multi_array():
    msgtype = "std_msgs/UInt8MultiArray"
    schema = dedent("""
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # Please look at the MultiArrayLayout message definition for
        # documentation on all multiarrays.

        MultiArrayLayout  layout        # specification of data layout
        uint8[]           data          # array of data
        ================================================================================
        MSG: std_msgs/MultiArrayLayout
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        # The multiarray declares a generic multi-dimensional array of a
        # particular data type.  Dimensions are ordered from outer most
        # to inner most.
        #
        # Accessors should ALWAYS be written in terms of dimension stride
        # and specified outer-most dimension first.
        #
        # multiarray(i,j,k) = data[data_offset + dim_stride[1]*i + dim_stride[2]*j + k]
        #
        # A standard, 3-channel 640x480 image with interleaved color channels
        # would be specified as:
        #
        # dim[0].label  = "height"
        # dim[0].size   = 480
        # dim[0].stride = 3*640*480 = 921600  (note dim[0] stride is just size of image)
        # dim[1].label  = "width"
        # dim[1].size   = 640
        # dim[1].stride = 3*640 = 1920
        # dim[2].label  = "channel"
        # dim[2].size   = 3
        # dim[2].stride = 3
        #
        # multiarray(i,j,k) refers to the ith row, jth column, and kth channel.

        MultiArrayDimension[] dim # Array of dimension properties
        uint32 data_offset        # padding bytes at front of data
        ================================================================================
        MSG: std_msgs/MultiArrayDimension
        # This was originally provided as an example message.
        # It is deprecated as of Foxy
        # It is recommended to create your own semantically meaningful message.
        # However if you would like to continue using this please use the equivalent in example_msgs.

        string label   # label of given dimension
        uint32 size    # size of given dimension (in type units)
        uint32 stride  # stride of given dimension
    """)

    with TemporaryDirectory() as temp_dir:
        msg = {
            "layout": {
                "dim": [{"label": "x", "size": 3, "stride": 3}],
                "data_offset": 0
            },
            "data": [0, 128, 255]
        }
        path = _write_mcap(temp_dir, msg, msgtype, schema)
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/rosbags"))

    assert len(messages) == 1
    assert messages[0].log_time == 0
    assert messages[0].publish_time == 0
    assert messages[0].sequence == 0
    assert messages[0].channel_id == 1
    assert len(messages[0].data.layout.dim) == 1
    assert messages[0].data.layout.dim[0].label == "x"
    assert messages[0].data.layout.dim[0].size == 3
    assert messages[0].data.layout.dim[0].stride == 3
    assert messages[0].data.layout.data_offset == 0
    assert list(messages[0].data.data) == [0, 128, 255]



import zlib
from abc import ABC, abstractmethod
from enum import IntEnum
from pathlib import Path


class FilePosition(IntEnum):
    START = 0    # Start of the file
    CURRENT = 1  # Current file position
    END = 2      # End of the file


class BaseReader(ABC):
    @abstractmethod
    def peek(self, size: int) -> bytes:
        """Peek at the next bytes in the reader."""
        ...  # pragma: no cover

    @abstractmethod
    def read(self, size: int | None = None) -> bytes:
        """Read the next bytes in the reader."""
        ...  # pragma: no cover

    @abstractmethod
    def seek_from_start(self, offset: int) -> int:
        """Seek from the start of the reader."""
        ...  # pragma: no cover

    @abstractmethod
    def seek_from_end(self, offset: int) -> int:
        """Seek from the end of the reader."""
        ...  # pragma: no cover

    @abstractmethod
    def seek_from_current(self, offset: int) -> int:
        """Seek from the current position of the reader."""
        ...  # pragma: no cover

    @abstractmethod
    def tell(self) -> int:
        """Get the current position in the reader."""
        ...  # pragma: no cover

    @abstractmethod
    def close(self) -> None:
        """Close the reader and release all resources."""
        ...  # pragma: no cover


class FileReader(BaseReader):
    def __init__(self, file_path: Path | str, mode: str = 'rb'):
        self._file_path = Path(file_path).absolute()
        self._file = open(self._file_path, mode)

    def peek(self, size: int) -> bytes:
        # Returns empty bytes when end of file
        return self._file.peek(size)

    def read(self, size: int | None = None) -> bytes:
        return self._file.read(size)

    def seek_from_start(self, offset: int) -> int:
        return self._file.seek(offset, FilePosition.START)

    def seek_from_end(self, offset: int) -> int:
        return self._file.seek(-offset, FilePosition.END)

    def seek_from_current(self, offset: int) -> int:
        return self._file.seek(offset, FilePosition.CURRENT)

    def tell(self) -> int:
        return self._file.tell()

    def close(self) -> None:
        self._file.close()


class BytesReader(BaseReader):
    def __init__(self, data: bytes):
        self._data = data
        self._position = 0

    def peek(self, size: int) -> bytes:
        # Returns empty bytes when end of data
        return self._data[self._position:self._position + size]

    def read(self, size: int | None = None) -> bytes:
        if size is None:
            return self._data
        result = self._data[self._position:self._position + size]
        self._position += size
        return result

    def seek_from_start(self, offset: int) -> int:
        self._position = offset
        return self._position

    def seek_from_end(self, offset: int) -> int:
        self._position = len(self._data) - offset
        return self._position

    def seek_from_current(self, offset: int) -> int:
        self._position += offset
        return self._position

    def tell(self) -> int:
        return self._position

    def align(self, size: int) -> 'BytesReader':
        if self._position % size > 0:
            self._position += size - (self._position % size)
        return self

    def size(self) -> int:
        return len(self._data)

    def close(self) -> None:
        pass


class CrcReader(BaseReader):
    def __init__(self, reader: BaseReader):
        self._reader = reader
        self._crc = 0

    def peek(self, size: int) -> bytes:
        return self._reader.peek(size)

    def read(self, size: int | None = None) -> bytes:
        data = self._reader.read(size)
        self._crc = zlib.crc32(data, self._crc)
        return data

    def seek_from_start(self, offset: int) -> int:
        return self._reader.seek_from_start(offset)

    def seek_from_end(self, offset: int) -> int:
        return self._reader.seek_from_end(offset)

    def seek_from_current(self, offset: int) -> int:
        return self._reader.seek_from_current(offset)

    def tell(self) -> int:
        return self._reader.tell()

    def close(self) -> None:
        self._reader.close()

    def get_crc(self) -> int:
        return self._crc

    def clear_crc(self) -> None:
        self._crc = 0



from pathlib import Path

from pybag.cli.main import main as cli_main
from pybag.mcap_writer import McapFileWriter
from pybag.ros2.humble.std_msgs import Int32


def test_cli_info_basic(tmp_path: Path, capsys) -> None:
    """Test basic info command output."""
    input_path = tmp_path / "input.mcap"

    # Create a simple MCAP with messages
    with McapFileWriter.open(input_path, chunk_size=1024) as writer:
        writer.write_message("/foo", int(1e9), Int32(data=1))
        writer.write_message("/bar", int(2e9), Int32(data=2))
        writer.write_message("/foo", int(3e9), Int32(data=3))

    # Run info command
    cli_main(["info", str(input_path)])

    # Capture output
    captured = capsys.readouterr()
    output = captured.out

    # Verify output contains expected information
    assert "input.mcap" in output
    assert "/foo" in output
    assert "/bar" in output



import zlib
from abc import ABC, abstractmethod
from pathlib import Path


class BaseWriter(ABC):
    """Abstract base class for binary writers."""

    @abstractmethod
    def write(self, data: bytes) -> int:
        """Write bytes."""
        ...  # pragma: no cover

    @abstractmethod
    def tell(self) -> int:
        """Get the current position in the writer."""
        ...  # pragma: no cover

    @abstractmethod
    def close(self) -> None:
        """Close the writer."""
        ...  # pragma: no cover


class FileWriter(BaseWriter):
    """Write binary data to a file."""

    def __init__(self, file_path: Path | str, mode: str = "wb"):
        self._file_path = Path(file_path).absolute()
        self._file = open(self._file_path, mode)
        self._bytes_written = 0

    def write(self, data: bytes) -> int:
        self._bytes_written += len(data)
        return self._file.write(data)

    def tell(self) -> int:
        return self._bytes_written

    def close(self) -> None:
        self._file.close()


class BytesWriter(BaseWriter):
    """Write binary data to an in-memory bytes buffer."""

    def __init__(self):
        self._buffer = bytearray()

    def write(self, data: bytes) -> int:
        self._buffer.extend(data)
        return len(data)

    def tell(self) -> int:
        return len(self._buffer)

    def align(self, size: int) -> None:
        current_length = len(self._buffer)
        if current_length % size > 0:
            padding = size - (current_length % size)
            self._buffer.extend(b"\x00" * padding)

    def size(self) -> int:
        return len(self._buffer)

    def as_bytes(self) -> bytes:
        return bytes(self._buffer)

    def clear(self) -> None:
        self._buffer.clear()

    def close(self) -> None:
        self._buffer.clear()


class CrcWriter(BaseWriter):
    """Write binary data and track CRC32."""

    def __init__(self, writer: BaseWriter):
        self._writer = writer
        self._crc = 0

    def write(self, data: bytes) -> int:
        self._crc = zlib.crc32(data, self._crc)
        return self._writer.write(data)

    def tell(self) -> int:
        return self._writer.tell()

    def get_crc(self) -> int:
        return self._crc

    def clear_crc(self) -> None:
        self._crc = 0

    def close(self) -> None:
        self._writer.close()



import ast
import dataclasses
import logging
import re
from dataclasses import fields, is_dataclass
from typing import Annotated, Any, Literal, get_args, get_origin

from pybag.io.raw_writer import BytesWriter
from pybag.mcap.records import SchemaRecord
from pybag.schema import (
    PRIMITIVE_TYPE_MAP,
    STRING_TYPE_MAP,
    Array,
    Complex,
    Primitive,
    Schema,
    SchemaConstant,
    SchemaDecoder,
    SchemaEncoder,
    SchemaEntry,
    SchemaField,
    SchemaFieldType,
    Sequence,
    String
)
from pybag.types import Message

logger = logging.getLogger(__name__)


class Ros2MsgError(Exception):
    """Exception raised for errors in the ROS2 message parsing."""
    def __init__(self, message: str):
        super().__init__(message)


class Ros2MsgSchemaDecoder(SchemaDecoder):
    def __init__(self):
        self._cache: dict[int, tuple[Schema, dict[str, Schema]]] = {}
        self._builtin_schemas = self._create_builtin_schemas()

    def _create_builtin_schemas(self) -> dict[str, Schema]:
        """Create schemas for built-in ROS2 message types."""
        builtin_schemas = {}

        # builtin_interfaces/Time
        builtin_schemas['builtin_interfaces/Time'] = Schema(
            'builtin_interfaces/Time',
            {
                'sec': SchemaField(Primitive('int32'), None),
                'nanosec': SchemaField(Primitive('uint32'), None)
            }
        )

        # builtin_interfaces/Duration
        builtin_schemas['builtin_interfaces/Duration'] = Schema(
            'builtin_interfaces/Duration',
            {
                'sec': SchemaField(Primitive('int32'), None),
                'nanosec': SchemaField(Primitive('uint32'), None)
            }
        )

        return builtin_schemas

    def _remove_inline_comment(self, line: str) -> str:
        in_single = False
        in_double = False
        for i, ch in enumerate(line):
            if ch == "'" and not in_double:
                in_single = not in_single
            elif ch == '"' and not in_single:
                in_double = not in_double
            elif ch == '#' and not in_single and not in_double:
                return line[:i].rstrip()
        return line.strip()

    def _parse_value(self, field_type: SchemaFieldType, raw_value: str) -> Any:
        if isinstance(field_type, Primitive):
            return PRIMITIVE_TYPE_MAP[field_type.type](raw_value)

        if isinstance(field_type, String):
            if field_type.max_length is not None and len(raw_value) > field_type.max_length:
                raise Ros2MsgError('String default exceeds length')
            return raw_value.strip('"') if raw_value.startswith('"') else raw_value.strip("'")

        if isinstance(field_type, (Array, Sequence)):
            values = ast.literal_eval(raw_value.strip())
            if not isinstance(values, list):
                raise Ros2MsgError('Array default must be a list')
            element_type = field_type.type
            if isinstance(element_type, Primitive):
                return [PRIMITIVE_TYPE_MAP[element_type.type](v) for v in values]
            else:
                raise Ros2MsgError('Default values not supported for this field type')

        raise Ros2MsgError('Default values not supported for this field type')

    def _parse_field_type(self, field_raw_type: str, package_name: str) -> SchemaFieldType:
        # Handle array and sequence types
        if array_match := re.match(r'(.*)\[(.*)\]$', field_raw_type):
            element_raw, length_spec = array_match.groups()
            element_field = self._parse_field_type(element_raw, package_name)

            if length_spec == '':
                return Sequence(element_field)

            is_bounded = length_spec.startswith('<=')
            length = int(length_spec[2:]) if is_bounded else int(length_spec)
            return Array(element_field, length, is_bounded)

        # Handle string types
        if field_raw_type.startswith('string'):
            if not (match := re.match(r'string(?:<=(\d+))?$', field_raw_type)):
                raise Ros2MsgError(f'Invalid string field: {field_raw_type}')
            length = int(match.group(1)) if match.group(1) else None
            return String('string', max_length=length)
        if field_raw_type.startswith('wstring'):
            if not (match := re.match(r'wstring(?:<=(\d+))?$', field_raw_type)):
                raise Ros2MsgError(f'Invalid wstring field: {field_raw_type}')
            length = int(match.group(1)) if match.group(1) else None
            return String('wstring', max_length=length)

        # Handle primitive types
        if field_raw_type in PRIMITIVE_TYPE_MAP:
            return Primitive(field_raw_type)

        # Handle complex types
        if field_raw_type == 'Header':
            field_raw_type = 'std_msgs/Header'
        elif '/' not in field_raw_type:
            field_raw_type = f'{package_name}/{field_raw_type}'
        return Complex(field_raw_type)

    def _parse_field(self, field: str, package_name: str) -> tuple[str, SchemaEntry]:
        # Extract the field type, name and optional default value
        if not (match := re.match(r'(\S+)\s+(\S+)(?:\s+(.+))?$', field)):
            raise Ros2MsgError(f'Invalid field definition: {field}')
        field_raw_type, field_raw_name, raw_default = match.groups()

        if not field_raw_type:
            raise Ros2MsgError('Field type cannot be empty')
        if not field_raw_name:
            raise Ros2MsgError('Field name cannot be empty')

        if is_constant := ('=' in field_raw_name or (raw_default and raw_default.startswith('='))):
            if '=' in field_raw_name:
                field_raw_name, raw_default = field_raw_name.split('=', 1)
            else:
                # TODO: Hack should be made more robust
                # Handle case where = is separated by spaces: "CONST = value"
                raw_default = raw_default[1:].strip()  # Remove the '=' and strip spaces
            if not field_raw_name.isupper():
                raise Ros2MsgError('Constant name must be uppercase')
            if not raw_default:
                raise Ros2MsgError('Constant value cannot be empty')
        else:
            if not re.match(r'[a-z][a-z0-9_]*$', field_raw_name):
                raise Ros2MsgError('Field name must be lowercase, alphanumeric or _')

        # Check the field name is valid
        if '__' in field_raw_name:
            raise Ros2MsgError('Field name cannot contain double underscore "__"')
        if field_raw_name.endswith('_'):
            raise Ros2MsgError('Field name cannot end with "_"')

        # Parse the field type
        schema_type = self._parse_field_type(field_raw_type, package_name)

        # Parse the default value if it exists
        default_value = None
        if raw_default is not None:
            # Certain field types cannot have default values
            if isinstance(schema_type, Complex):
                raise Ros2MsgError('Complex fields cannot have default values')
            if isinstance(schema_type, Array) and isinstance(schema_type.type, String):
                raise Ros2MsgError('Array of strings cannot have default values')
            if isinstance(schema_type, Sequence) and isinstance(schema_type.type, String):
                raise Ros2MsgError('Sequence of strings cannot have default values')
            default_value = self._parse_value(schema_type, raw_default)

        if is_constant:
            if default_value is None:
                raise Ros2MsgError('Constant must have a default value')
            return field_raw_name, SchemaConstant(schema_type, default_value)
        return field_raw_name, SchemaField(schema_type, default_value)

    def _add_missing_builtin_schemas(
        self,
        main_schema: SchemaRecord,
        sub_schemas: dict[str, Schema]
    ) -> None:
        """Add any missing built-in schemas that are referenced but not defined."""
        schema_text = main_schema.data.decode('utf-8')
        for builtin_name, builtin_schema in self._builtin_schemas.items():
            if builtin_name not in sub_schemas and builtin_name in schema_text:
                sub_schemas[builtin_name] = builtin_schema

    def parse_schema(self, schema: SchemaRecord) -> tuple[Schema, dict[str, Schema]]:
        if schema.id in self._cache:
            return self._cache[schema.id]

        assert schema.encoding == "ros2msg"

        package_name = schema.name.split('/')[0]
        msg = schema.data.decode('utf-8')

        # Remove comments and empty lines
        lines = [self._remove_inline_comment(line) for line in msg.split('\n')]
        lines = [line for line in lines if line]
        msg = '\n'.join(lines)

        # Split along '=' delimiter
        msg = [m.strip() for m in msg.split('=' * 80)]

        msg_schema = {}
        # The first message does not have the 'MSG: ' prefix line
        main_fields = [m.strip() for m in msg[0].split('\n') if m.strip()]
        for raw_field in main_fields:
            field_name, field = self._parse_field(raw_field, package_name)
            msg_schema[field_name] = field

        sub_msg_schemas = {}
        for sub_msg in msg[1:]:
            sub_msg_name = sub_msg.split('\n')[0].strip()[5:]  # Remove 'MSG: ' prefix
            sub_msg_fields = [m.strip() for m in sub_msg.split('\n')[1:] if m]
            # TODO: Do some caching here
            sub_msg_schema = {}
            # Use the package name from the sub-message, not the main message
            sub_msg_package_name = sub_msg_name.split('/')[0]
            for raw_field in sub_msg_fields:
                field_name, field = self._parse_field(raw_field, sub_msg_package_name)
                sub_msg_schema[field_name] = field
            sub_msg_schemas[sub_msg_name] = Schema(sub_msg_name, sub_msg_schema)

        # Add any required built-in schemas
        main_schema = Schema(schema.name, msg_schema)
        self._add_missing_builtin_schemas(schema, sub_msg_schemas)
        result = main_schema, sub_msg_schemas

        self._cache[schema.id] = result
        return result


class Ros2MsgSchemaEncoder(SchemaEncoder):
    def __init__(self):
        self._cache = None  # TODO: Cache messages we come across

    @classmethod
    def encoding(cls) -> str:
        return "ros2msg"

    def _extract_literal_int(self, literal_type: Any) -> int:
        """Extract the value from a Literal type annotation."""
        if hasattr(literal_type, '__origin__') and literal_type.__origin__ is Literal:
            if literal_args := get_args(literal_type):
                return int(literal_args[0])
            else:
                raise Ros2MsgError(f"Unknown literal type: {literal_type}")
        if isinstance(literal_type, str) and literal_type.isdigit():
            return int(literal_type)
        if isinstance(literal_type, int):
            return literal_type
        raise Ros2MsgError(f"Unknown literal type: {literal_type}")

    def _parse_annotation(self, annotation_type: Any) -> SchemaFieldType:
        annotation_args = get_args(annotation_type)
        if len(annotation_args) < 2:
            raise Ros2MsgError(f"Field is not correctly annotated.")

        field_type = annotation_args[-1]
        if field_type[0] in PRIMITIVE_TYPE_MAP:
            return Primitive(field_type[0])

        if field_type[0] in STRING_TYPE_MAP:
            return String(field_type[0])

        if field_type[0] == 'array':
            sub_type = self._parse_annotation(field_type[1])
            if (length := field_type[2]) is None:
                return Sequence(sub_type)
            # Extract the actual integer value from Literal types
            actual_length = self._extract_literal_int(length)
            return Array(sub_type, length=actual_length, is_bounded=False)

        if field_type[0] == 'complex':
            return Complex(field_type[1])

        if field_type[0] == 'constant':
            return self._parse_annotation(field_type[1])

        raise Ros2MsgError(f"Unknown field type: {field_type}")

    def _parse_default_value(self, annotation: dataclasses.Field) -> Any:
        if annotation.default is not dataclasses.MISSING:
            return annotation.default
        if annotation.default_factory is not dataclasses.MISSING:
            return annotation.default_factory()
        return None

    def _parse_message(self, message: Message | type[Message]) -> tuple[Schema, dict[str, Schema]]:
        if not is_dataclass(message):
            raise TypeError("Expected a dataclass instance")

        cls = message if isinstance(message, type) else type(message)
        class_name = cls.__msg_name__

        schema = Schema(class_name, {})
        sub_schemas: dict[str, Schema] = {}

        for field in fields(cls):
            if get_origin(field.type) is not Annotated:
                raise Ros2MsgError(f"Field '{field.name}' is not correctly annotated.")

            field_type = self._parse_annotation(field.type)
            field_default = self._parse_default_value(field)

            if get_args(field.type)[-1][0] == 'constant':
                schema.fields[field.name] = SchemaConstant(field_type, field_default)
                continue
            schema.fields[field.name] = SchemaField(field_type, field_default)

            if isinstance(field_type, Sequence):
                if isinstance(field_type.type, Complex):
                    # For Sequence[Complex[Class]], extract the Class from the annotation
                    list_type = get_args(field.type)[0]  # list[Annotated[Class, ...]]
                    if get_origin(list_type) is list and get_args(list_type):
                        complex_annotation = get_args(list_type)[0]  # Annotated[Class, ...]
                        if get_origin(complex_annotation) is Annotated:
                            complex_type = get_args(complex_annotation)[0]  # Class
                            sub_schema, sub_sub_schemas = self._parse_message(complex_type)
                            sub_schemas[sub_schema.name] = sub_schema
                            sub_schemas.update(sub_sub_schemas)

            if isinstance(field_type, Array):
                if isinstance(field_type.type, Complex):
                    # For Array[Complex[Class]], extract the Class from the annotation
                    list_type = get_args(field.type)[0]  # list[Annotated[Class, ...]]
                    if get_origin(list_type) is list and get_args(list_type):
                        complex_annotation = get_args(list_type)[0]  # Annotated[Class, ...]
                        if get_origin(complex_annotation) is Annotated:
                            complex_type = get_args(complex_annotation)[0]  # Class
                            sub_schema, sub_sub_schemas = self._parse_message(complex_type)
                            sub_schemas[sub_schema.name] = sub_schema
                            sub_schemas.update(sub_sub_schemas)

            if isinstance(field_type, Complex):
                complex_type = get_args(field.type)[0]
                sub_schema, sub_sub_schemas = self._parse_message(complex_type)
                sub_schemas[sub_schema.name] = sub_schema
                sub_schemas.update(sub_sub_schemas)

        return schema, sub_schemas

    def _type_str(self, field_type: SchemaFieldType) -> str:
        if isinstance(field_type, Primitive):
            return field_type.type
        if isinstance(field_type, String):
            if field_type.max_length is None:
                return field_type.type
            return f'{field_type.type}<={field_type.max_length}'
        if isinstance(field_type, Array):
            sub_type = self._type_str(field_type.type)
            if field_type.is_bounded:
                return f'{sub_type}[<={field_type.length}]'
            return f'{sub_type}[{field_type.length}]'
        if isinstance(field_type, Sequence):
            return f'{self._type_str(field_type.type)}[]'
        if isinstance(field_type, Complex):
            return field_type.type.replace('/msg/', '/')
        raise Ros2MsgError(f'Unknown field type: {field_type}')

    def _value_str(self, value: Any) -> str:
        if isinstance(value, bool):
            return 'true' if value else 'false'
        if isinstance(value, (int, float, str, bytes)):
            return str(value)
        if isinstance(value, list):
            return f'[{", ".join(self._value_str(v) for v in value)}]'
        raise Ros2MsgError(f'Unknown value type: {type(value)}')

    def _encode_constant(self, writer: BytesWriter, field_name: str, field: SchemaConstant) -> None:
        encoded_type = self._type_str(field.type)
        encoded_name = field_name.upper()
        encoded_value = self._value_str(field.value)
        writer.write(f'{encoded_type} {encoded_name}={encoded_value}\n'.encode('utf-8'))

    def _encode_field(self, writer: BytesWriter, field_name: str, field: SchemaField) -> None:
        encoded_type = self._type_str(field.type)
        if field.default is not None:
            encoded_value = self._value_str(field.default)
            writer.write(f'{encoded_type} {field_name} {encoded_value}\n'.encode('utf-8'))
        else:
            writer.write(f'{encoded_type} {field_name}\n'.encode('utf-8'))

    def encode(self, message: Message | type[Message]) -> bytes:
        schema, sub_schemas = self._parse_message(message)

        writer = BytesWriter()
        for field_name, field in schema.fields.items():
            if isinstance(field, SchemaConstant):
                self._encode_constant(writer, field_name, field)
            elif isinstance(field, SchemaField):
                self._encode_field(writer, field_name, field)

        for sub_schema in sub_schemas.values():
            writer.write(('=' * 80 + '\n').encode('utf-8'))
            writer.write(f'MSG: {sub_schema.name.replace("/msg/", "/")}\n'.encode('utf-8'))
            for field_name, field in sub_schema.fields.items():
                if isinstance(field, SchemaConstant):
                    self._encode_constant(writer, field_name, field)
                elif isinstance(field, SchemaField):
                    self._encode_field(writer, field_name, field)

        return writer.as_bytes()

    def parse_schema(self, message: Message | type[Message]) -> tuple[Schema, dict[str, Schema]]:
        return self._parse_message(message)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    pose_schema = SchemaRecord(
        id=1,
        name="geometry_msgs/msg/Pose",
        encoding='ros2msg',
        data=b'# A representation of pose in free space, composed of position and orientation.\n\nPoint position\nQuaternion orientation\n\n================================================================================\nMSG: geometry_msgs/Point\n# This contains the position of a point in free space\nfloat64 x\nfloat64 y\nfloat64 z\n\n================================================================================\nMSG: geometry_msgs/Quaternion\n# This represents an orientation in free space in quaternion form.\n\nfloat64 x 0\nfloat64 y 0\nfloat64 z 0\nfloat64 w 1\n'
    )
    point_schema = SchemaRecord(
        id=1,
        name='geometry_msgs/msg/Point',
        encoding='ros2msg',
        data=b'# This contains the position of a point in free space\nfloat64 x\nfloat64 y\nfloat64 z\n'
    )
    header_schema = SchemaRecord(
        id=1,
        name='std_msgs/msg/Header',
        encoding='ros2msg',
        data=b'builtin_interfaces/Time stamp\nstring frame_id\n================================================================================\nMSG: builtin_interfaces/Time\nint32 sec\nuint32 nanosec\n'
    )
    pose_with_covariance_schema = SchemaRecord(
        id=1,
        name='geometry_msgs/msg/PoseWithCovariance',
        encoding='ros2msg',
        data=b'geometry_msgs/Pose pose\nfloat64[36] covariance\n================================================================================\nMSG: geometry_msgs/Pose\ngeometry_msgs/Point position\ngeometry_msgs/Quaternion orientation\n================================================================================\nMSG: geometry_msgs/Point\nfloat64 x\nfloat64 y\nfloat64 z\n================================================================================\nMSG: geometry_msgs/Quaternion\nfloat64 x\nfloat64 y\nfloat64 z\nfloat64 w\n'
    )
    diagnostic_status_schema = SchemaRecord(
        id=1,
        name='diagnostic_msgs/msg/DiagnosticStatus',
        encoding='ros2msg',
        data=b'byte OK=0\nbyte WARN=1\nbyte ERROR=2\nbyte STALE=3\nbyte level\nstring name\nstring message\nstring hardware_id\ndiagnostic_msgs/KeyValue[] values\n================================================================================\nMSG: diagnostic_msgs/KeyValue\nstring key\nstring value\n'
    )



"""MCAP filtering CLI command."""

import argparse
import fnmatch
import logging
from collections import defaultdict
from pathlib import Path
from textwrap import dedent
from typing import Literal

from pybag.io.raw_writer import FileWriter
from pybag.mcap.record_reader import McapRecordReaderFactory
from pybag.mcap.record_writer import McapRecordWriterFactory
from pybag.mcap.records import MessageRecord

logger = logging.getLogger(__name__)


def _to_ns(seconds: float | None) -> int | None:
    if seconds is None:
        return None
    return int(seconds * 1_000_000_000)


def filter_mcap(
    input_path: str | Path,
    output_path: str | Path | None = None,
    include_topics: list[str] | None = None,
    exclude_topics: list[str] | None = None,
    start_time: float | None = None,
    end_time: float | None = None,
    chunk_size: int | None = None,
    chunk_compression: Literal["lz4", "zstd"] | None = None,
    *,
    overwrite: bool = False
) -> Path:
    """Filter an MCAP file based on topics and time."""
    # Resolve input and output paths
    input_path = Path(input_path).resolve()
    if output_path is None:
        output_path = input_path.with_name(f"{input_path.stem}_filtered.mcap")
    output_path = Path(output_path).resolve()
    if output_path == input_path:
        raise ValueError('Input path cannot be same as output.')

    # Check if output path exists
    if not overwrite and output_path.exists():
        raise ValueError('Output mcap exists. Please set `overwrite` to True.')

    start_ns, end_ns = _to_ns(start_time), _to_ns(end_time)
    with McapRecordReaderFactory.from_file(input_path) as reader:
        # Build topic -> channel_ids mapping
        all_channels = reader.get_channels()
        topic_to_channel_ids: dict[str, set[int]] = defaultdict(set)
        for channel_id, channel in all_channels.items():
            topic_to_channel_ids[channel.topic].add(channel_id)
        all_topics = set(topic_to_channel_ids.keys())

        # Determine which topics to include
        if include_topics is None:  # Include all topics
            topics_to_include = set(topic_to_channel_ids.keys())
        else:  # Expand glob patterns and filter
            topics_to_include = set()
            for pattern in include_topics:
                matched = fnmatch.filter(all_topics, pattern)
                topics_to_include.update(matched)

        # Remove excluded topics
        if exclude_topics:
            topics_to_exclude = set()
            for pattern in exclude_topics:
                matched = fnmatch.filter(all_topics, pattern)
                topics_to_exclude.update(matched)
            topics_to_include -= topics_to_exclude

        channel_ids_to_include = set()
        for topic in topics_to_include:
            channel_ids_to_include.update(topic_to_channel_ids[topic])

        # Step 2: Write the filtered MCAP using factory
        with McapRecordWriterFactory.create_writer(
            FileWriter(output_path),
            chunk_size=chunk_size,
            chunk_compression=chunk_compression,
            profile=reader.get_header().profile
        ) as writer:
            # Write message records as we read them
            written_schema_ids = set()
            written_channel_ids = set()
            sequence_counters = defaultdict(int)

            # If no topics match the filters, create an empty MCAP (no messages)
            if not channel_ids_to_include:
                logger.warning("No topics match filter.")
                return output_path

            for msg_record in reader.get_messages(
                channel_id=list(channel_ids_to_include),
                start_timestamp=start_ns,
                end_timestamp=end_ns,
                in_log_time_order=False
            ):
                # Write the schema record to the mcap the first time (if it has one)
                # Note: schema_id == 0 means "no schema" and is valid in MCAP
                schema_id = all_channels[msg_record.channel_id].schema_id
                if schema_id != 0 and schema_id not in written_schema_ids:
                    if (schema := reader.get_schema(schema_id)) is not None:
                        writer.write_schema(schema)
                        written_schema_ids.add(schema_id)
                    else:
                        channel_id = msg_record.channel_id
                        logger.warning(f'Schema {schema_id} not found for channel {channel_id}')

                # Write the channel record to the mcap the first time
                if msg_record.channel_id not in written_channel_ids:
                    writer.write_channel(all_channels[msg_record.channel_id])
                    written_channel_ids.add(msg_record.channel_id)

                # Write message immediately with updated sequence number
                new_record = MessageRecord(
                    channel_id=msg_record.channel_id,
                    sequence=sequence_counters[msg_record.channel_id],
                    log_time=msg_record.log_time,
                    publish_time=msg_record.publish_time,
                    data=msg_record.data
                )
                sequence_counters[msg_record.channel_id] += 1
                writer.write_message(new_record)

    return output_path


def add_parser(subparsers) -> None:
    parser = subparsers.add_parser(
        "filter",
        help="Extract data from an mcap.",
        description=dedent("""
            The messages that satisfy the given constraints are written into
            a new mcap file. The new mcap is written from scratch, so the way
            the messages are stored in the file compared to the input mcap may
            be different (e.g messages may no longer be in the same chunk).
        """)
    )
    parser.add_argument("input", help="Path to mcap file (*.mcap).")
    parser.add_argument("-o", "--output", help="Output MCAP file path")
    parser.add_argument(
        "--include-topic",
        action="append",
        help=dedent("""
            Topics to include. If not specified, defaults to all topics.
            If specified, only the topics listed are included.
            Excluded topics are ignored (i.e. exclusion takes precedent).
        """)
    )
    parser.add_argument(
        "--exclude-topic",
        action="append",
        help=dedent("""
            Topics to exclude. If not specified, defaults to no topics.
            If specified, the specified topics are excluded from the output.
        """)
    )
    parser.add_argument(
        "--start-time",
        type=float,
        help=dedent("""
            Start time in seconds. All messages with a log time less than
            the start time is ignored and not included in the output mcap.
            By default it is set to the smallest log time in the input mcap.
        """)
    )
    parser.add_argument(
        "--end-time",
        type=float,
        help=dedent("""
            End time in seconds. All messages with a log time greater than
            the start time is ignored and not included in the output mcap.
            By default it is set to the largest log time in the input mcap.
        """)
    )
    parser.add_argument(
        "--chunk-size",
        type=int,
        help=dedent("""Chunk size of the new filtered mcap in bytes.""")
    )
    parser.add_argument(
        "--chunk-compression",
        type=str,
        choices=['lz4', 'zstd'],
        help=dedent("""Compression used for the chunk records.""")
    )
    parser.set_defaults(
        func=lambda args: filter_mcap(
            args.input,
            output_path=args.output,
            include_topics=args.include_topic,
            exclude_topics=args.exclude_topic,
            start_time=args.start_time,
            end_time=args.end_time,
            chunk_size=args.chunk_size,
            chunk_compression=args.chunk_compression,
        )
    )



from collections import deque
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Iterator

from mcap.reader import make_reader
from mcap_ros2.decoder import DecoderFactory
from pytest_benchmark.fixture import BenchmarkFixture
from rosbags.highlevel import AnyReader
from rosbags.typesys import Stores, get_typestore

from pybag.mcap_reader import McapFileReader

from .benchmark_utils import create_test_mcap


def read_with_pybag(mcap: Path) -> Iterator[Any]:
    with McapFileReader.from_file(mcap) as reader:
        for topic in reader.get_topics():
            for message in reader.messages(topic):
                yield message.data


def read_with_rosbags(mcap: Path) -> Iterator[Any]:
    typestore = get_typestore(Stores.LATEST)
    with AnyReader([mcap.parent]) as reader:
        for conn, _, data in reader.messages():
            yield typestore.deserialize_cdr(data, conn.msgtype)


def read_with_official(mcap: Path) -> Iterator[Any]:
    with open(mcap, "rb") as f:
        reader = make_reader(f, decoder_factories=[DecoderFactory()])
        for _, _, _, ros_msg in reader.iter_decoded_messages(log_time_order=False):
            yield ros_msg


def test_official_read(benchmark: BenchmarkFixture) -> None:
    with TemporaryDirectory() as tmpdir:
        mcap = create_test_mcap(Path(tmpdir) / "test")
        benchmark(lambda: deque(read_with_official(mcap), maxlen=0))


def test_rosbags_read(benchmark: BenchmarkFixture) -> None:
    with TemporaryDirectory() as tmpdir:
        mcap = create_test_mcap(Path(tmpdir) / "test")
        benchmark(lambda: deque(read_with_rosbags(mcap), maxlen=0))


def test_pybag_read(benchmark: BenchmarkFixture) -> None:
    with TemporaryDirectory() as tmpdir:
        mcap = create_test_mcap(Path(tmpdir) / "test")
        benchmark(lambda: deque(read_with_pybag(mcap), maxlen=0))



from dataclasses import dataclass
from enum import IntEnum

from pybag.crc import assert_crc
from pybag.mcap.error import McapUnknownCompressionError


@dataclass
class HeaderRecord:
    profile: str
    library: str


@dataclass
class FooterRecord:
    summary_start: int
    summary_offset_start: int
    summary_crc: int


@dataclass
class ChannelRecord:
    id: int
    schema_id: int
    topic: str
    message_encoding: str
    metadata: dict[str, str]


@dataclass
class SchemaRecord:
    id: int
    name: str
    encoding: str
    data: bytes


@dataclass
class MessageRecord:
    channel_id: int
    sequence: int
    log_time: int
    publish_time: int
    data: bytes


@dataclass
class ChunkRecord:
    message_start_time: int
    message_end_time: int
    uncompressed_size: int
    uncompressed_crc: int
    compression: str
    records: bytes


@dataclass
class MessageIndexRecord:
    channel_id: int
    records: list[tuple[int, int]]


@dataclass
class ChunkIndexRecord:
    message_start_time: int
    message_end_time: int
    chunk_start_offset: int
    chunk_length: int
    message_index_offsets: dict[int, int]
    message_index_length: int
    compression: str
    compressed_size: int
    uncompressed_size: int


@dataclass
class AttachmentRecord:
    log_time: int
    create_time: int
    name: str
    media_type: str
    data: bytes
    crc: int


@dataclass
class MetadataRecord:
    name: str
    metadata: dict[str, str]


@dataclass
class DataEndRecord:
    data_section_crc: int


@dataclass
class AttachmentIndexRecord:
    offset: int
    length: int
    log_time: int
    create_time: int
    data_size: int
    name: str
    media_type: str


@dataclass
class MetadataIndexRecord:
    offset: int
    length: int
    name: str


@dataclass
class StatisticsRecord:
    message_count: int
    schema_count: int
    channel_count: int
    attachment_count: int
    metadata_count: int
    chunk_count: int
    message_start_time: int
    message_end_time: int
    channel_message_counts: dict[int, int]


@dataclass
class SummaryOffsetRecord:
    group_opcode: int
    group_start: int
    group_length: int


class RecordType(IntEnum):
    MAGIC_BYTES = 0x89
    HEADER = 0x01
    FOOTER = 0x02
    SCHEMA = 0x03
    CHANNEL = 0x04
    MESSAGE = 0x05
    CHUNK = 0x06
    MESSAGE_INDEX = 0x07
    CHUNK_INDEX = 0x08
    ATTACHMENT = 0x09
    METADATA = 0x0C
    DATA_END = 0x0F
    ATTACHMENT_INDEX = 0x0A
    METADATA_INDEX = 0x0D
    STATISTICS = 0x0B
    SUMMARY_OFFSET = 0x0E


def decompress_chunk(chunk: ChunkRecord, *, check_crc: bool = False) -> bytes:
    """Decompress the records field of a chunk."""
    if chunk.compression == 'zstd':
        import zstandard as zstd
        chunk_data = zstd.ZstdDecompressor().decompress(chunk.records)
    elif chunk.compression == 'lz4':
        import lz4.frame
        chunk_data = lz4.frame.decompress(chunk.records)
    elif chunk.compression == '':
        chunk_data = chunk.records
    else:
        error_msg = f'Unknown compression type: {chunk.compression}'
        raise McapUnknownCompressionError(error_msg)

    # Validate the CRC if requested
    if check_crc and chunk.uncompressed_crc != 0:
        assert_crc(chunk_data, chunk.uncompressed_crc)
    return chunk_data
